/***************************************************************************
* ======================================================================
* Computer Vision/Image Processing Tool Project - Dr. Scott Umbaugh SIUE
* ======================================================================
*
*             File Name: assemble_bands.c
*           Description: assemble bands from several monochrome images to
*			 produce a multi-spectral image.  Return the
*			 "assembled" image.
*         Related Files: libband.a
*   Initial Coding Date: 8/12/96
*           Portability: Standard (ANSI) C
*             Credit(s): Arve Kjoelen
*                        Southern Illinois University @ Edwardsville
*
** Copyright (C) 1996 SIUE - by Scott Umbaugh and Arve Kjoelen.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
**
****************************************************************************/

#include "stdlib.h"
#include "stdio.h"
#include "CVIPimage.h"


/****************************************************************
                   Name: assemble_band
          Expanded Name: Assemble Image Bans
             Parameters: <inImgs> - pointer to pointers to image structures.
			 <noimgs> - number of images passed.
			 a monochrome image.
                Returns: pointer to new Image structure containing assembled
			 image.
            Description: assemble a multiband image from the passed monochrome
			 images.
	    Diagnostics: - returns with NULL pointer and error message when
			 assembly cannot be performed.
			 - If n images are passed, the result image will have
			 n bands.
			 - If any of the passed images are not monochrome,
			 the first spectral band of these images will be
			 used for asdsembly.  The other bands will be ignored.
			 - Support for complex images:  If all input images
			   are COMPLEX, the output will also be COMPLEX.  If
			   one or more of the input images are REAL, the output
			   will be REAL.
			 - The input Image structures are not deleted.
              Credit(s): Arve Kjoelen
****************************************************************/

Image *assemble_bands(Image **inImgs, int noimgs) {
   Image *outIm;
   HISTORY stuff;
   int *hptr, cnt;
   FORMAT fm=COMPLEX;
   const char fn[] = {"assemble_bands"};
   
  /*
   * Some sanity checks first...
   */
   if(inImgs == (Image **)NULL) {
      fprintf(stderr, "%s: No input Images!\n",fn);
      return((Image *)NULL);
   }
   for(cnt=0; cnt<noimgs; cnt++) {
      if(inImgs[cnt] == (Image *)NULL ) {
         fprintf(stderr, "%s: Too few input images %d. %d needed\n",fn,cnt+1,noimgs);
         return((Image *)NULL);
      }
      if(inImgs[cnt]->image_ptr[0]->data_format==REAL) fm=REAL;
      if((inImgs[cnt]->image_ptr[0]->rows != inImgs[0]->image_ptr[0]->rows) || (inImgs[cnt]->image_ptr[0]->cols != inImgs[0]->image_ptr[0]->cols)) {
         fprintf(stderr, "%s: Input images must be of identical size.  Try resizing your images.\n",fn);
         return((Image *)NULL);
      }
   }
   
  /*
   * Allocate memory and set new image parameters.
   */
   outIm=(Image *)malloc(sizeof(Image));
   outIm->bands=noimgs; 
   outIm->story=(HISTORY)NULL;
   if(noimgs==3) {
      outIm->image_format=PPM;
      outIm->color_space=RGB;
   }
   else {
      outIm->image_format=VIP;
      outIm->color_space=GRAY_SCALE;
   }

  /*
   * Assemble the image and delete input Image header structures.
   */
   outIm->image_ptr=(Matrix **)malloc(sizeof(Matrix *)*noimgs);
   for(cnt=0; cnt<noimgs; cnt++) {
      if(inImgs[cnt]->bands > 1) {
         fprintf(stderr, "%s: Warning: input image # %d of %d\n",fn,cnt,noimgs);
         fprintf(stderr, "%s: has > one band.  Ignoring all but band 0\n",fn);
      }
      outIm->image_ptr[cnt]=duplicate_Matrix(inImgs[cnt]->image_ptr[0], fm);
   }
   
  /*
   * Add new history information.
   */
   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=1;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_INTEGER;
   hptr=(int *)malloc(sizeof(int));
   hptr[0]=noimgs;
   stuff->packetP->dptr=(void **)malloc(sizeof(void *));
   stuff->packetP->dptr[0]=(int *)hptr;
   stuff->next=(HISTORY)NULL;
   stuff->ftag=ASSEMBLE_BANDS;
   history_add(outIm, stuff);

   return(outIm);
}
      
