/************************************************************************
*									*
* ====================================================================  *
*  Computer Vision/Image Processing Tool - Dr. Scott Umbaugh, SIUE      *
* ====================================================================  *
*									*
*	FILE NAME: vq.c							*
*									*
*	DESCRIPTION: This file contains routines necessary to implement *
*	"standard" vector quantization, based on block.  The routines   *
*	herein utilize the data structures defined in LBG.h.		* 
*									*
*	AUTHOR: Arve Kjoelen, SIUE					*
*									*
*	RELATED FILES: LBG.h, LBG.c					*
*									*
*	INITIAL CODING DATE: 9/27/93					*
*									*
*	LAST MODIFICATIONS MADE: 12/30/93				*
*									*
*									*
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "CVIPtoolkit.h"
#include "CVIPdef.h"
#include "CVIPtransform.h"
#include "LBG.h"

/************************************************************************
*									*
*	ROUTINE/PGM NAME:	wvq_setup()				*
*									*
*	DESCRIPTION:		This setup routine for wavelet-based	*
*		vector quantization lets the user choose between 	*
*		several 'canned' compression schemes.  These 		*
*		shcemes take care of everything - wavelet transform,	*
*		vector quantization and output to disk.  The input is 	*
*		the image we want to compress.				*
*									*
*	AUTHOR: Arve Kjoelen, SIUE	(akjoele@siue.edu)		*
*									*
*	INITIAL CODING DATE: 5/11/95					*
*									*
*	LAST MODIFICATIONS MADE: Thu May 11 11:56:01 CDT 1995		*
*									*
************************************************************************/
Image *wvq_setup(Image *image) {
	char *answer=malloc(32);
	char *name;
	int arg1, arg2, **hptr, *tptr, choice, done=0, maxsz, size, j, dim, i=0;
        int hsize, vsize, nblocks, nvecs, lowband, fbands;
	HISTORY stuff;
	VQ_INFO info, temp;	/* Vector Quantization info structure */
	ELEMENT struc;		/* Structure related to Vector Quantization */
	CVIP_BOOLEAN remap;

	if(image->image_ptr[0]->rows != image->image_ptr[0]->cols) {
		maxsz=MAX(image->image_ptr[0]->rows, image->image_ptr[0]->cols);
		while(size>1) {
			if(size>maxsz)
				size>>=1;
			else
				(void)zero_pad(image, 2*size, 2*size);
		}
	}
	fprintf(stdout, "\nVector Quantization operates on equal-sized\n");
	fprintf(stdout, "image blocks.  Please select the blocksize you\n");
	fprintf(stdout, "wish to use for your image.  Remember that the\n");
	fprintf(stdout, "blocksize should be divisible into the image size\n");
	while (!done) {
		fprintf(stdout, "Enter horizontal blocksize:");
		(void)gets(answer);
		hsize=atoi(answer);
		if(hsize>0 && !((image->image_ptr[0]->cols)%hsize)) {
			fprintf(stdout, "\nEnter vertical blocksize:");
			(void)gets(answer);
			vsize=atoi(answer);
			if(vsize>0 && !((image->image_ptr[0]->rows)%vsize)) 
				done=1;
			else 
				fprintf(stderr, "Invalid Blocksize\n");
		}	
		else
			fprintf(stderr, "Invalid Blocksize\n");
	}
	nvecs=(image->image_ptr[0]->rows*image->image_ptr[0]->cols)/(hsize*vsize);
	fprintf(stdout, "Please enter the number of codebook vectors you");
	fprintf(stdout, "wish to use (max=%d).  This number should\n", nvecs);
	fprintf(stdout, "be an integer power of two\n");
	done=0;
	while(!done) {
		fprintf(stdout, "Enter number of codebook vectors:\n");
		(void)gets(answer);
		choice=atoi(answer);
		if((check_pow_two(choice))==CVIP_YES)
			done=1;
		else {
			fprintf(stderr, "Your number is not a power of two\n");
		}
	}
	free(answer);
	fprintf(stdout, "Your Compression Ratio will be %f\n", (float)(nvecs/choice));
	/* 
	 * Right here, we need a routine which does the structure
	 * setup, fills in the parameters, etc.
	 */
	if(image->image_ptr[0]->data_type !=CVIP_FLOAT)
      		(void) cast_Image(image, CVIP_FLOAT);

	lowband=(int)(image->image_ptr[0]->rows)>>2;
	image=(Image *)wavdaub4_transform(image, 1, lowband);

	/* Initialization stuff for Vector Quantization */
	fprintf(stdout, "Performing Vector Quantization...\n");

   /* Allocate main structure for vector quantization */
   info= (VQ_INFO)malloc(sizeof(struct VQ_info));
   info->block_type=EQUAL_BLOCKS;
   info->pct_done=CVIP_NO;
   info->elements=(ELEMENT)calloc(1, sizeof(struct element));
   temp=(VQ_INFO)info;
   while((++i)<image->bands) {
      temp->next = (VQ_INFO)calloc(1, sizeof(struct VQ_info));
      temp->next->block_type=EQUAL_BLOCKS;
      temp->next->pct_done=CVIP_NO;
      temp->next->elements=(ELEMENT)calloc(1, sizeof(struct element));
      temp=(VQ_INFO)temp->next;
   }
   temp->next = (VQ_INFO)NULL;
   temp=info;
   switch (choice) {
      case 1:
	 fbands=10;
	 info->scheme=7;
	 break;
      case 2:
	 fbands=10;
	 info->scheme=8;
	 break;
      case 3:
	 fbands=10;
	 info->scheme=2; 
	 break;
      case 4:
         fbands=10;
         info->scheme=5;
         break;	
      default:
	 fprintf(stderr, "Error, invalid scheme specified\n");
	 return(Image *)NULL;
   }
   get_lbg_scheme((int)info->scheme, image->bands, info);
   allocate_vectors(image, info);
   for(i=0; i<image->bands; i++) {
      if(image->bands>1) fprintf(stdout, "Color band #%d\n", i);
      for(j=0, struc = (ELEMENT)info->elements; j<info->bands; j++, struc=(ELEMENT)struc->next) {
      if(info->codebook_size[j]) {
         dim=(info->hsize[j]*info->vsize[j]);
         fprintf(stderr, "Iterating, frequency band #%d: Rel. Improv.:", j);
         done=LBG_error(info->codebook_size[j], dim, struc, struc->codebook, 1, info->nvecs[j]);
         encode_data(struc->codebook, struc->data, struc->codebook_index,
info->codebook_size[j], dim, info->nvecs[j]);
         compute_centroid(struc, info->codebook_size[j], dim, info->nvecs[j]);
         while(!LBG_error(info->codebook_size[j], dim, struc, struc->codebook, 0, info->nvecs[j])) {
            encode_data(struc->codebook, struc->data, struc->codebook_index, info->codebook_size[j], dim, info->nvecs[j]);
            compute_centroid(struc, info->codebook_size[j], dim, info->nvecs[j]);
            }
         }
      }   
      info=info->next;
   }
   fprintf(stdout, "Encoding complete.\n");
   info=temp;
   name=save_codebook(image, info);
   lbg_save(image, info, name);

   /* Free data structures */
   free_info(info, image->bands);
   info=(VQ_INFO)NULL;
   
   /* Allocate main structure for vector quantization */
   i=0;
   info= (VQ_INFO)malloc(sizeof(struct VQ_info));
   info->pct_done=CVIP_NO;
   info->block_type=WAVL;
   info->elements=(ELEMENT)calloc(1, sizeof(struct element));
   temp=(VQ_INFO)info;
   while((++i)<image->bands) {
      temp->next = (VQ_INFO)calloc(1, sizeof(struct VQ_info));
      temp->next->block_type=WAVL;
      temp->next->elements=(ELEMENT)calloc(1, sizeof(struct element));
      temp=(VQ_INFO)temp->next;
   }
   temp->next = (VQ_INFO)NULL;

   delete_Image(image);
   image=(Image *)NULL;

   fprintf(stderr, "Loading codebook: %s\n", name);
   if(!(load_lbg_book(info, name))) {
      fprintf(stderr, "Error reading codebook\n");
      return(Image *)NULL;
   }

   image=lbg_decompress(info, name);
   do_decoding(image, info);
   image=wav_setup(image, &remap);
/*
   free_info(info, image->bands);
   stuff=history_get(image, WAVELET);
   hptr=(int **)stuff->packetP->dptr;
   arg1=hptr[0][0];
   arg2=hptr[1][0];
   switch(hptr[0][0]) {
      case 1: image=(Image *)wavhaar_transform(image, -1, hptr[1][0]);
              break;
      case 2: image=(Image *)wavdaub4_transform(image, -1, hptr[1][0]);
              break;
      default: fprintf(stderr, "wvq_setup(): Warning: Error detected in history structure\n");
	      break;
   }
*/
   return(image);
}
