/************************************************************************
*									*
* ====================================================================  *
*  Computer Vision/Image Processing Tool - Dr. Scott Umbaugh, SIUE      *
* ====================================================================  *
*									*
*	FILE NAME: vq.c						*
*									*
*	SYNOPSIS: This file contains routines related to the LBG	*
*	(Linde, Buzo, Gray) algorithm for vector quantization.  Among	*
*	the routines are algorithms that remap square blocks of an 	*
*	image to one-dimensional arrays for use as input to the LBG	*
*	algorithm.  Routines that implement the recursive LBG method	*
*	for Vector Quantization are also in this file.			*
*									*
*	AUTHOR: Arve Kjoelen, SIUE					*
*									*
*	RELATED FILES: LBG.h, LBG.c					*
*									*
*	INITIAL CODING DATE: 9/27/93					*
*									*
*	LAST MODIFICATIONS MADE: 12/30/93				*
*									*
*									*
************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "CVIPtoolkit.h"
#include "CVIPdef.h"
#include "LBG.h"
#include "wavelet.h"

/************************************************************************
*									*
*	ROUTINE/PGM NAME:	wvq_setup()				*
*									*
*	DESCRIPTION:		This setup routine for wavelet-based	*
*		vector quantization lets the user choose between 	*
*		several 'canned' compression schemes.  These 		*
*		shcemes take care of everything - wavelet transform,	*
*		vector quantization and output to disk.  The input is 	*
*		the image we want to compress.				*
*									*
*	AUTHOR: Arve Kjoelen, SIUE	(akjoele@siue.edu)		*
*									*
*	INITIAL CODING DATE: 5/11/95					*
*									*
*	LAST MODIFICATIONS MADE: Thu May 11 11:56:01 CDT 1995		*
*									*
************************************************************************/
Image *wvq_setup(Image *image) {
	char *answer=malloc(32);
	char *name;
	int arg1, arg2, **hptr, *tptr, choice, done, lowband, maxsz, size, fbands, j, dim, i=0;
	HISTORY stuff;
	VQ_INFO info, temp;	/* Vector Quantization info structure */
	ELEMENT struc;		/* Structure related to Vecotr Quantization */
	CVIP_BOOLEAN remap;

	if(image->image_ptr[0]->rows != image->image_ptr[0]->cols) {
		maxsz=MAX(image->image_ptr[0]->rows, image->image_ptr[0]->cols);
		while(size>1) {
			if(size>maxsz)
				size>>=1;
			else
				(void)zero_pad(image, 2*size, 2*size);
		}
	}
	fprintf(stdout, "\nAVAILABLE COMPRESSION SCHEMES:\n");
	fprintf(stdout, "1.	10.16:1	Compression - slow\n");
	fprintf(stdout, "2.	15:1 	Compression - slow\n");
	fprintf(stdout, "3.	32.87:1	Compression - slow\n");
	fprintf(stdout, "4.	35.73:1 Compression - fair speed\n\n");
	fprintf(stdout, "Please Enter Your Choice:\n");
	(void)gets(answer);
	choice=atoi(answer);
	free(answer);

	if(image->image_ptr[0]->data_type !=CVIP_FLOAT)
      		(void) cast_Image(image, CVIP_FLOAT);

	/* Initialization stuff for wavelet transform */
	fprintf(stdout, "Performing Wavelet Xform:  10 subbands, Wavelet=daub4...\n");
	lowband=(int)(image->image_ptr[0]->rows)>>2;
	image=(Image *)wavdaub4_transform(image, 1, lowband);

	/* Add history information for wavelet */
	stuff=(HISTORY)NULL;
   	stuff=(HISTORY)malloc(sizeof(struct history));
   	stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   	stuff->packetP->dsize=2;
   	stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE)*2);
   	stuff->packetP->dtype[0]=CVIP_INTEGER;
   	stuff->packetP->dtype[1]=CVIP_INTEGER;
   	hptr=(int **)NULL;
   	hptr=(int **)malloc(sizeof(int *)*2);
	hptr[0]=(int *)malloc(sizeof(int));
	hptr[1]=(int *)malloc(sizeof(int));
   	hptr[0][0]=2; /* 1=haar, 2=daub4 			*/
	if(choice<5)
   		hptr[1][0]=3; /* 1=4 fbands, 2=7fbands, 3=10fbands */
	else
		fprintf(stdout, "Not implemented\n");

   	stuff->packetP->dptr=(void **)hptr;
   	stuff->next=(HISTORY)NULL;
   	stuff->ftag=WAVELET;
   	history_add(image, stuff);


	/* Initialization stuff for Vector Quantization */
	fprintf(stdout, "Performing Vector Quantization...\n");

   /* Allocate main structure for vector quantization */
   info= (VQ_INFO)malloc(sizeof(struct VQ_info));
   info->block_type=WAVL;
   info->pct_done=CVIP_NO;
   info->elements=(ELEMENT)calloc(1, sizeof(struct element));
   temp=(VQ_INFO)info;
   while((++i)<image->bands) {
      temp->next = (VQ_INFO)calloc(1, sizeof(struct VQ_info));
      temp->next->block_type=WAVL;
      temp->next->pct_done=CVIP_NO;
      temp->next->elements=(ELEMENT)calloc(1, sizeof(struct element));
      temp=(VQ_INFO)temp->next;
   }
   temp->next = (VQ_INFO)NULL;
   temp=info;
   switch (choice) {
      case 1:
	 fbands=10;
	 info->scheme=7;
	 break;
      case 2:
	 fbands=10;
	 info->scheme=8;
	 break;
      case 3:
	 fbands=10;
	 info->scheme=2; 
	 break;
      case 4:
         fbands=10;
         info->scheme=5;
         break;	
      default:
	 fprintf(stderr, "Error, invalid scheme specified\n");
	 return(Image *)NULL;
   }
   get_lbg_scheme((int)info->scheme, image->bands, info);
   allocate_vectors(image, info);
   for(i=0; i<image->bands; i++) {
      if(image->bands>1) fprintf(stdout, "Color band #%d\n", i);
      for(j=0, struc = (ELEMENT)info->elements; j<info->bands; j++, struc=(ELEMENT)struc->next) {
      if(info->codebook_size[j]) {
         dim=(info->hsize[j]*info->vsize[j]);
         fprintf(stderr, "Iterating, frequency band #%d: Rel. Improv.:", j);
         done=LBG_error(info->codebook_size[j], dim, struc, struc->codebook, 1, info->nvecs[j]);
         encode_data(struc->codebook, struc->data, struc->codebook_index,
info->codebook_size[j], dim, info->nvecs[j]);
         compute_centroid(struc, info->codebook_size[j], dim, info->nvecs[j]);
         while(!LBG_error(info->codebook_size[j], dim, struc, struc->codebook, 0, info->nvecs[j])) {
            encode_data(struc->codebook, struc->data, struc->codebook_index, info->codebook_size[j], dim, info->nvecs[j]);
            compute_centroid(struc, info->codebook_size[j], dim, info->nvecs[j]);
            }
         }
      }   
      info=info->next;
   }
   fprintf(stdout, "Encoding complete.\n");
   info=temp;
   name=save_codebook(image, info);
   lbg_save(image, info, name);

   /* Free data structures */
   free_info(info, image->bands);
   info=(VQ_INFO)NULL;
   
   /* Allocate main structure for vector quantization */
   i=0;
   info= (VQ_INFO)malloc(sizeof(struct VQ_info));
   info->pct_done=CVIP_NO;
   info->block_type=WAVL;
   info->elements=(ELEMENT)calloc(1, sizeof(struct element));
   temp=(VQ_INFO)info;
   while((++i)<image->bands) {
      temp->next = (VQ_INFO)calloc(1, sizeof(struct VQ_info));
      temp->next->block_type=WAVL;
      temp->next->elements=(ELEMENT)calloc(1, sizeof(struct element));
      temp=(VQ_INFO)temp->next;
   }
   temp->next = (VQ_INFO)NULL;

   delete_Image(image);
   image=(Image *)NULL;

   fprintf(stderr, "Loading codebook: %s\n", name);
   if(!(load_lbg_book(info, name))) {
      fprintf(stderr, "Error reading codebook\n");
      return(Image *)NULL;
   }

   image=lbg_decompress(info, name);
   do_decoding(image, info);
   image=wav_setup(image, &remap);
/*
   free_info(info, image->bands);
   stuff=history_get(image, WAVELET);
   hptr=(int **)stuff->packetP->dptr;
   arg1=hptr[0][0];
   arg2=hptr[1][0];
   switch(hptr[0][0]) {
      case 1: image=(Image *)wavhaar_transform(image, -1, hptr[1][0]);
              break;
      case 2: image=(Image *)wavdaub4_transform(image, -1, hptr[1][0]);
              break;
      default: fprintf(stderr, "wvq_setup(): Warning: Error detected in history structure\n");
	      break;
   }
*/
   return(image);
}
