/***************************************************************************
======================================================================
Computer Vision/Image Processing Tool Project - Dr. Scott Umbaugh SIUE
======================================================================
                                                                      
             File Name: m_makecomplex.c
           Description: contains routine for reallocating space to a 
			real MATRIX structure to make it complex
         Related Files: Makefile, CVIPmatrix.h
   Initial Coding Date: 1/24/93
           Portability: Standard (ANSI) C
             Credit(s): Gregory Hance, E.E. Dept. SIUE

  Global Function List: m_makecomplex

			
****************************************************************************

** Copyright (C) 1991, 1992, 1993 SIUE - by Scott Umbaugh, Greg Hance.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.

RCS (Revision Control System) Information - Added automatically by RCS) 

$Author: yawei $                                                         
$Date: 1997/03/08 00:55:42 $                                                        
$Source: /export/home_faculty/cvipapps/CVIPtools3.5/MatrixAlgebra/RCS/m_makecomplex.c,v $                                                         
$Header: /export/home_faculty/cvipapps/CVIPtools3.5/MatrixAlgebra/RCS/m_makecomplex.c,v 1.4 1997/03/08 00:55:42 yawei Exp $

*****************************************************************************/
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "CVIPmatrix.h"


/****************************************************************
                  Name: matrix_makecomplex
         Expanded Name: Make a Complex Matrix
            Parameters: <real_MATRIX> - pointer to real matrix
			<verbose> - speak to me
               Returns: 1 - upon success, 0 - upon failure
           Description: reallocates twice the amount of memory
			already assigned to <real_MATRIX>  to make a complex.
			The new complex data can be referenced
			through **iptr of the MATRIX structure.
			realloc() is used to keep the new data address'
			contiguous with that of the real data...
			rather than allocate a new block of memory.
			This has potential advantages when peforming
			the same operation on both data sets (real & imag.)
           Diagnostics: returns a NULL pointer and warning message when...
			   ...matrix is already complex
			   ...memory cannot be successfully allocated
             Credit(s): Greg Hance
			
			
****************************************************************/
CVIP_BOOLEAN matrix_makecomplex(real_MATRIX, verbose)
    MATRIX *real_MATRIX;
    CVIP_BOOLEAN verbose;
{
    int i, j, rows, cols;
    CVIP_TYPE data_type;
    char **data;
    const char function_name[] = {"matrix_makecomplex"};
    
/* check to make sure matrix is already not a complex MATRIX */
    if(real_MATRIX->data_format == COMPLEX) {
       fprintf(stderr, "%s: the matrix is already complex goofy!\n",function_name);
       return(0);
    }
    
/* change matrix format to that of complex MATRIX */
    real_MATRIX->data_format = COMPLEX;

/* simplify some commonly used variables */    
    rows = real_MATRIX->rows;
    cols = real_MATRIX->cols;
    data_type = real_MATRIX->data_type;
    data = real_MATRIX->rptr;
    
/* acknowledge complex allocation */
   if(verbose)
      fprintf(stderr, "%s: extending MATRIX to complex format.\n",function_name);

/* try to allocate the request */
    switch(data_type) {
        case CVIP_BYTE: {    /* byte matrix */
            byte **byte_matrix;
            if( !(byte_matrix = (byte **)realloc(data, 2*rows*sizeof(byte *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*byte_matrix = (byte *)realloc(*data, 2*rows*cols*sizeof(byte))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

            /* remap the table of pointers in making it a 2-d complex array */
	    /* since a pointer to a (possibly new) memory region may be returned */
	    /* by realloc() we must remap the real data as well as the imaginary. */

            for(i = cols; i < 2*cols*rows; i += cols)
               byte_matrix[i/cols] = (byte *) *byte_matrix + i;

            /* initialize all of the imaginary data elements to value 0 */    
	    (void) bzero(byte_matrix[rows],cols*rows);

            real_MATRIX->rptr = (char **)byte_matrix;
	    real_MATRIX->iptr = (char **)byte_matrix + rows;
            break;
        }
        case CVIP_SHORT: {    /* short matrix */
            short **short_matrix;
            if( !(short_matrix = (short **)realloc(data, 2*rows*sizeof(short *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*short_matrix = (short *)realloc(*data, 2*cols*rows*sizeof(short))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

            /* remap the table of pointers in making it a 2-d array */

            for(i = cols; i < 2*cols*rows; i += cols)
               short_matrix[i/cols] = (short *) *short_matrix + i;
	       
            /* initialize all of the imaginary data elements to value 0 */    
	    (void) bzero(short_matrix[rows],cols*rows*sizeof(short));

            real_MATRIX->rptr = (char **)short_matrix;
	    real_MATRIX->iptr = (char **)short_matrix + rows;
            break;
        }
        case CVIP_INTEGER: {    /* integer matrix */
            int **integer_matrix;
            if( !(integer_matrix = (int **)realloc(data, 2*rows*sizeof(int *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*integer_matrix = (int *)realloc(*data, 2*cols*rows*sizeof(int))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

            /* remap the table of pointers in making it a 2-d array */

            for(i = cols; i < 2*cols*rows; i += cols)
               integer_matrix[i/cols] = (int *) *integer_matrix + i;
	       
            /* initialize all of the imaginary data elements to value 0 */    
	    (void) bzero(integer_matrix[rows],cols*rows*sizeof(int));

            real_MATRIX->rptr = (char **) integer_matrix;
	    real_MATRIX->iptr = (char **) integer_matrix + rows;
            break;
        }
        case CVIP_FLOAT: {    /* float matrix */
            float **float_matrix;
            if( !(float_matrix = (float **)realloc(data, 2*rows*sizeof(float *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*float_matrix = (float *) realloc(*data, 2*cols*rows*sizeof(float))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

            /* remap the table of pointers in making it a 2-d array */

            for(i = cols; i < 2*cols*rows; i += cols)
               float_matrix[i/cols] = (float *) *float_matrix + i;
	       
            /* initialize all of the imaginary data elements to value 0 */    
	    (void) bzero(float_matrix[rows],cols*rows*sizeof(float));

            real_MATRIX->rptr = (char**)float_matrix;
	    real_MATRIX->iptr = (char**)float_matrix + rows;
            break;
        }
        case CVIP_DOUBLE: {    /* double matrix */
            double **double_matrix;
            if( !(double_matrix = (double **)realloc(data, 2*rows*sizeof(double *))) ){
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*double_matrix = (double *)realloc(*data, 2*cols*rows*sizeof(double))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

            /* remap the table of pointers in making it a 2-d array */

            for(i = cols; i <2*cols*rows; i += cols)
               double_matrix[i/cols] = (double *) *double_matrix + i;
	 
            /* initialize all of the imaginary data elements to value 0 */    
	    (void) bzero(double_matrix[rows],cols*rows*sizeof(double));

            real_MATRIX->rptr = (char **)double_matrix;
	    real_MATRIX->iptr = (char **)double_matrix + rows;
            break;
        }
        default:
            printf("\nERROR in matrix_allocate: unsupported data_type\n");
            return(0);
    }
    return(1);
}
