
/***************************************************************************
======================================================================
Computer Vision/Image Processing Tool Project - Dr. Scott Umbaugh SIUE
======================================================================
                                                                      
             File Name: m_makereal.c
           Description: contains routine to truncate an imaginary
			matrix into a real
         Related Files: Makefile, CVIPmatrix.h
   Initial Coding Date: 1/24/93
           Portability: Standard (ANSI) C
             Credit(s): Gregory Hance, E.E. Dept. SIUE

  Global Function List: makeReal_Matrix - make a real matrix
			
****************************************************************************

** Copyright (C) 1991, 1992, 1993 SIUE - by Scott Umbaugh, Greg Hance.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.

RCS (Revision Control System) Information - Added automatically by RCS) 

$Author: yawei $                                                         
$Date: 1997/03/08 00:55:49 $                                                        
$Source: /export/home_faculty/cvipapps/CVIPtools3.5/MatrixAlgebra/RCS/m_makereal.c,v $                                                         
$Header: /export/home_faculty/cvipapps/CVIPtools3.5/MatrixAlgebra/RCS/m_makereal.c,v 1.3 1997/03/08 00:55:49 yawei Exp $

*****************************************************************************/
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "CVIPmatrix.h"

/****************************************************************
                  Name: makeReal_Matrix
         Expanded Name: Matrix Make Real
            Parameters: complex_Matrix - pointer to matrix structure
			verbose - should i speak?
               Returns: 1 - upon success, 0 - upon failure
           Description: reallocates the amount of memory associated 
			with complex_Matrix to that of a real matrix
           Diagnostics: returns NULL pointer and error message when...
				...matrix is already real
				...memory cannot be reallocated by realloc()
             Credit(s): Greg Hance
			
			
****************************************************************/
CVIP_BOOLEAN makeReal_Matrix(complex_Matrix, verbose)
    Matrix *complex_Matrix;
    CVIP_BOOLEAN verbose;
{
    int i, j, rows, cols;
    CVIP_TYPE data_type;
    char **data;
    const char function_name[] = {"makeReal_Matrix"};

/* check to make sure matrix is already not a real Matrix */
    if(complex_Matrix->data_format == REAL) {
       fprintf(stderr, "makeReal_Matrix: the matrix is already real goofy!\n");
       return(0);
    }

/* change matrix format to that of complex Matrix */
    complex_Matrix->data_format = REAL;
    
/* simplify some commonly used variables */    
    rows = complex_Matrix->rows;
    cols = complex_Matrix->cols;
    data_type = complex_Matrix->data_type;
    data = complex_Matrix->rptr;
      
/* acknowledge complex allocation */
   if(verbose)
      fprintf(stderr, "makeReal_Matrix: cropping complex part of Matrix to form real matrix.\n");
      
/* try to allocate the request */
    switch(data_type) {
        case CVIP_BYTE: {    /* byte matrix */
            byte **byte_matrix;
            if( !(byte_matrix = (byte **)realloc(data, rows*sizeof(byte *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*byte_matrix = (byte *)realloc(*data, rows*cols*sizeof(byte))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

	    /* Since a pointer to a (possibly new) memory region may be returned */
	    /*  by realloc() we may need to remap the real data. */
            if( ((char *)*byte_matrix) != *data )
               for(i = cols; i < cols*rows; i += cols)
                  byte_matrix[i/cols] = (byte *) *byte_matrix + i;

            complex_Matrix->rptr = (char **)byte_matrix;
            break;
        }
        case CVIP_SHORT: {    /* short matrix */
            short **short_matrix;
            if( !(short_matrix = (short **)realloc(data, rows*sizeof(short *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*short_matrix = (short *)realloc(*data, cols*rows*sizeof(short))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

	    /* Since a pointer to a (possibly new) memory region may be returned */
	    /* by realloc() we may need to remap the real data. */
            if( ((char *)*short_matrix) != *data)
               for(i = cols; i < cols*rows; i += cols)
                  short_matrix[i/cols] = (short *) *short_matrix + i;
	       
            complex_Matrix->rptr = (char **)short_matrix;
            break;
        }
        case CVIP_INTEGER: {    /* integer matrix */
            int **integer_matrix;
            if( !(integer_matrix = (int **)realloc(data, rows*sizeof(int *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*integer_matrix = (int *)realloc(*data, cols*rows*sizeof(int))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

	    /* Since a pointer to a (possibly new) memory region may be returned */
	    /* by realloc() we may need to remap the real data. */
            if( ((char *)*integer_matrix) != *data)
               for(i = cols; i < cols*rows; i += cols)
                  integer_matrix[i/cols] = (int *) *integer_matrix + i;

            complex_Matrix->rptr = (char **) integer_matrix;
            break;
        }
        case CVIP_FLOAT: {    /* float matrix */
            float **float_matrix;
            if( !(float_matrix = (float **)realloc(data, rows*sizeof(float *))) ) {
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*float_matrix = (float *) realloc(*data, cols*rows*sizeof(float))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

	    /* Since a pointer to a (possibly new) memory region may be returned */
	    /* by realloc() we may need to remap the real data. */
            if( ((char *)*float_matrix) != *data)
               for(i = cols; i < cols*rows; i += cols)
                  float_matrix[i/cols] = (float *) *float_matrix + i;
	       
            complex_Matrix->rptr = (char**)float_matrix;
            break;
        }
        case CVIP_DOUBLE: {    /* double matrix */
            double **double_matrix;
            if( !(double_matrix = (double **)realloc(data, rows*sizeof(double *))) ){
	       fprintf(stderr,"%s: could not reallocate first level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
	    }
	    if( !(*double_matrix = (double *)realloc(*data, cols*rows*sizeof(double))) ) {
	       fprintf(stderr,"%s: could not reallocate second level\n",function_name);
	       fprintf(stderr,"%s: of indirection for pointer variable **rptr.\n",function_name);
	       return(0);
            }

	    /* Since a pointer to a (possibly new) memory region may be returned */
	    /* by realloc() we may need to remap the real data. */
            if( ((char *)*double_matrix) != *data)
               for(i = cols; i <cols*rows; i += cols)
                  double_matrix[i/cols] = (double *) *double_matrix + i;
	 
            complex_Matrix->rptr = (char **)double_matrix;
            break;
        }
        default:
            printf("\nERROR in new_Matrix: unsupported data_type\n");
            return(0);
    }

complex_Matrix->iptr = NULL;
return(1);
}
