/***************************************************************************
* ======================================================================
* Computer Vision/Image Processing Tool Project - Dr. Scott Umbaugh SIUE
* ======================================================================
*
*             File Name: dermvis.c
*           Description: example program for CVIPtools that demonstrates
*                        command line processing and use of the image conversion
*                        utilities as well as simple use of an external function
*                        "threshold" used to process and image. Also handles cases
*                        for reading from "stdin" or "stdout".
*                        Some examples of program usage...
*			 (e.g) zcat pepper.pgm.Z | cvipskel | xv -
*                        	- reads an image file from "stdin" (uncompressing first)
*                        and processes the image sending the resultant image to "stdout"
*                        which is then piped to a display program.
*                        (e.g.) cvipskel -display pepper.pgm test.pgm
*			 	- reads the image from the file pepper.pgm, processes
*			 the image with the default threshold (128) and stores the
*			 resultant image out to test.pgm which is then displayed.
*                        (e.g.) cvipskel -thresh 0.7 pepper.pgm > test.pgm
*			 	- reads the image from the file pepper.pgm, processes
*			 the image w/ the specified threshold of (0.7*255 = 179) and
*                        sends the result to "stdout" which is redirected to the file
*                        test.pgm, but not displayed.
*                           One benefit demonstrated by the second example is that it 
*			 is possible to form a series of filters using pipes...
*                        For example say we have three programs...
*     			 1 - fft, 2- lowpass, 3- ifft which perform a forward FFT, lowpass
*                        filtering and inverse FFT respectively then the following series
*                        of filters could be used to process and display the image...
*                        fft pepper.pgm | lowpass | ifft | xv -
*         Related Files: threshold.c, threshold.h
*   Initial Coding Date: Wed Apr 14 14:02:34 CDT 1993
*           Portability: Standard (ANSI) C
*             Credit(s): Gregory Hance
*                        Southern Illinois University @ Edwardsville
*
* RCS (Revision Control System) Information
* ...Added automatically by RCS 
*
* $Log: DermVis.c,v $
* Revision 1.11  1997/03/08 17:01:20  yawei
* Swaped CVIP_YES and CVIP_NO
*
* Revision 1.10  1997/03/08 00:44:29  yawei
* Name Changes:
* 	BOOLEAN ==> CVIP_BOOLEAN
* 	FALSE ==> CVIP_NO
* 	TRUE ==> CVIP_YES
* 	BYTE ==> CVIP_BYTE
* 	SHORT ==> CVIP_SHORT
* 	INTEGER ==> CVIP_INTEGER
* 	FLOAT ==> CVIP_FLOAT
* 	DOUBLE ==> CVIP_DOUBLE
* 	TYPE ==> CVIP_TYPE
*
* Revision 1.9  1997/02/26 21:29:21  yawei
* link ==> slink for porting to Linux
*
* Revision 1.8  1996/12/31 00:25:19  yawei
* crop_Image ==> crop
*
 * Revision 1.7  1996/12/23  18:20:59  yawei
 * changed setup function names
 *
 * Revision 1.6  1996/12/23  00:45:49  yawei
 * Changed the setup function names
 *
 * Revision 1.5  1996/12/21  01:24:49  yawei
 * Changed function names of segmentations.
 *
 * Revision 1.4  1996/12/17  22:06:06  yawei
 * Changed "image_geometry.h" to "CVIPgeometry.h"
 *
 * Revision 1.3  1996/12/13  19:01:46  akjoele
 * included image_geometry.h
 *
 * Revision 1.2  1996/12/13  18:57:13  akjoele
 * changed call to crop_Image since this function now
 * returns an Image * instead of an int
 *
 * Revision 1.1  1994/10/30  17:58:25  hanceg
 * CVIPtools1.0 source file - G. Hance
 *
 * Revision 1.1  1993/04/29  06:55:50  hanceg
 * Initial revision
 *
*
****************************************************************************/
#include <ctype.h>
/* Define Preprocessor statements */
#include "CVIPdef.h"
#include "CVIPtoolkit.h"
#include "CVIPconvert.h"
#include "CVIPgeometry.h"
#include "DermVis.h"


/*
** global variables used for 1st class compatibility
*/
byte 		*maxP, *minP;
float		*aveP, *varP;
float		gmh_cog, gmv_cog;

Image *
dermVisApp_SetUp(
	const Image	*imgP
)
{
	CVIP_BOOLEAN		gray = CVIP_NO;
	Image		*borP, *maskP, *freP, *labP = NULL, *savP;
	Matrix		*matP = NULL;
	ObjectList 	listP;
	Object 		*objP;
	ChainCode 	*ccP;
	BorderPoints	 b_pnts; /* for spline only */
	Color 		bg = {0, 0, 0};
	char		buf[80], *fn = "dermVisApp", outputfile[80];
	int		ray_x, ray_y, *xcP=NULL, *ycP=NULL;
	unsigned	rows, cols, skip, choice;
	float		ksize;
	byte		*bP;

	borP = duplicate_Image(imgP);

	rows  = getNoOfRows_Image(borP);
	cols  = getNoOfCols_Image(borP);

	print_CVIP("\n");
	msg_CVIP(fn,"Do you wish to crop the image (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		msg_CVIP(fn, "Enter the number of rows to crop.\n");
		msg_CVIP(fn, "image has %u rows: ", rows);
		skip = getUInt_CVIP(10, 1, rows-10);
		msg_CVIP(fn, "cropping lower %u rows of image...\n", skip );
		borP=crop(borP, 0, 0, rows-skip, cols);
	}

	print_CVIP("\n");
	msg_CVIP(fn,"Do you wish to perform pseudo-median filtering (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		msg_CVIP(fn, "peforming pseudo-median filtering ...\n");
		(void) PMedian_Image(borP);
	}

	print_CVIP("\n");
	msg_CVIP(fn,"Do you wish to perform nonskin masking (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		msg_CVIP(fn, "peforming nonskin masking ...\n");
		maskP = findNonSkin_DVis(borP);
		freP = borP;
		borP = mask_Image(borP, maskP);
		delete_Image(freP);
	}

	print_CVIP("\n");
	msg_CVIP(fn,"Please select a segmentation method:\n\n");
	msg_CVIP(fn,"(1) histogram thresholding\n");
	msg_CVIP(fn,"(2) split and merge\n");
	msg_CVIP(fn,"(3) PCT/median cut\n");
	msg_CVIP(fn,"(4) SCT/center split\n");
	msg_CVIP(fn,"(5) multiresolution (enter 0 for level)\n");
	msg_CVIP(fn,"(6) fuzzy c-mean\n\n");
	msg_CVIP(fn,"Your choice: ");
	choice = getUInt_CVIP(10, 1, 6);
	switch(choice) {
		case 1:	borP = hist_thresh_segment(borP); break;
		case 2:	borP = split_merge_setup(borP); break; 
		case 3:	borP = median_cut_setup(borP); gray = CVIP_YES; break;
		case 4:	borP = sct_split_setup(borP); gray = CVIP_YES; break;
		case 5:	borP = split_merge_setup(borP); break; 
		case 6:	borP = fuzzyc_segment_setup(borP); gray = CVIP_YES; break;
		default: error_CVIP(fn, "bad case - shouldn't happen!\n");
	}
	if(!borP) {
		error_CVIP(fn, "error during segmentation routine!");
		return NULL;
	}

	print_CVIP("\n");
	msg_CVIP(fn,"Do you wish to perform morphological data reduction (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		print_CVIP("\n");
		msg_CVIP(fn, "Please enter the kernel dimension.\n");
		msg_CVIP(fn, "(e.g.) 3 would result in a 3x3 kernel matrix.\n");
		msg_CVIP(fn, "Your choice: ");
		ksize = (float) getFloat_CVIP(3, 9);

		if(gray) {
			freP = borP;
			borP = color2index_Image(borP);
			delete_Image(freP);
		}

		if(!disk_SE(&matP, CVIP_INTEGER, -1.0, 0.0, ksize)) {
			error_CVIP(fn,"could not form SE");
			return NULL;
		}

		print_CVIP("\n");
		if(!MorphOpen_Image(borP, matP, 0, 0, 0)) {
			error_CVIP(fn,
			"could not perform gray-scale morphological opening");
			return NULL;
		}

		print_CVIP("\n");
		if(!MorphClose_Image(borP, matP, 0, 0, 0)) {
			error_CVIP(fn,
			"could not perform gray-scale morphological closing");
			return NULL;
		}

		delete_Matrix(matP);
	}

	print_CVIP("\n\n");
	msg_CVIP(fn,"Do you wish to save the image at this point (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		msg_CVIP(fn, "Please enter the file name to be saved under: ");
		(void) fGetLine_CVIP(stdin, outputfile);
		savP = remap_Image(borP, CVIP_BYTE, 0, 255);
		setFileFormat_Image(savP, PGM);
		write_Image(savP, outputfile, CVIP_NO, CVIP_NO, PGM, 1);
		delete_Image(savP);
		display_Image(outputfile, PGM);
	}


	print_CVIP("\n");
	listP = label_Objects( borP, &labP, 255);
	delete_Image(borP);
	borP = labP;

	print_CVIP("\n");
	getProp_Objects(listP, borP);

	trimList_Objects(listP, getNoOfCols_Image(borP), 
	getNoOfRows_Image(borP), 0.0050, 0.9, 0.1, 1.0, -90.0,
	90.0, 0.5, 0.5, CVIP_YES);

	if(!size_LL(listP)) {
		freP = borP;
		borP = new_Border(getNoOfCols_Image(borP),
		getNoOfRows_Image(borP));
		getBandVector_Image(borP, &bP);
		(void) memset((void *)bP, 0x01,
		getNoOfCols_Image(borP)*getNoOfRows_Image(borP));
	}

	print_CVIP("\n");
	msg_CVIP(fn, "Enter the centroid for which the object closest to\n");
	msg_CVIP(fn, "this point will be selected as the border object.\n");
	msg_CVIP(fn, "Horizontal\tCoord.: ");
	gmh_cog = getUInt_CVIP(10, 0, cols-1);
	msg_CVIP(fn, "Vertical\tCoord.: ");
 	gmv_cog = getUInt_CVIP(10, 0, rows-1);


	if(!getBest_Objects(listP)) {
		error_CVIP(fn, "allocation failure!");
		return NULL;
	}

	borP = 	new_Border(getNoOfCols_Image(borP),getNoOfRows_Image(borP));

	head_LL(listP);

	print_CVIP("\n");
	for (;;) {

		objP = retrieve_LL(listP);
		getLocDat_Image(labP, objP->label, &ray_x, &ray_y);
		ccP = new_ChainCode(ray_y,ray_x,objP->label);
		build_ChainCode(ccP, labP, objP->x_min, objP->y_min, 	
		objP->x_max, objP->y_max);
		draw_ChainCode(ccP, borP);
		getXY_ChainCode(ccP, &xcP, &ycP);
		b_pnts.xP = subSample_Vector(xcP, ccP->no_of_vectors,
		CVIP_INTEGER, 15, &b_pnts.no_of_points);
		b_pnts.yP = subSample_Vector(ycP, ccP->no_of_vectors, 	
		CVIP_INTEGER, 15, &b_pnts.no_of_points);
		b_pnts.h_span = getNoOfCols_Image(borP);
		b_pnts.v_span = getNoOfRows_Image(borP);
		free(xcP);
		free(ycP);

		if (istail_LL(listP)) break;
		next_LL(listP);
	}
	delete_Image(labP);

	print_CVIP("\n");
	msg_CVIP(fn,"Do you wish to spline the border (y/n) [y]?");
	(void) fGetLine_CVIP(stdin, buf);
	if(tolower(*buf) == 'y' || *buf == '\0') {
		freP = borP;
   		borP = new_Border(b_pnts.h_span, b_pnts.v_span);
   		drawSpline_Border(borP, b_pnts.no_of_points, b_pnts.xP, b_pnts.yP);
   		delete_Image(freP);
	}

	getInsidePoint_Contour(borP, 1, &ray_x, &ray_y);
   	fillContour_Border(borP, 1, ray_x+1, ray_y+3,  
   	objP->x_min, objP->y_min, objP->x_max, objP->y_max);

	return borP;
}



void 
fillContour_Border(
	Border imageP,
	int nv,
	int x,
	int y,
	int x_min,
	int y_min,
	int x_max,
	int y_max
)
{

	Window *regionP;
	byte **dP;

        regionP = (Window *) malloc(sizeof(Window));

        regionP->x0 = x_min;
        regionP->x1 = x_max;
        regionP->y0 = y_min;
        regionP->y1 = y_max;

	dP = (byte **) getData_Matrix(getBand_Image(imageP,0));
        fill( dP, x, y, regionP, nv);
}

Border
new_Border(
	unsigned width,
	unsigned height
)
{
	return( new_Image( PBM, BINARY, 1, height, width, CVIP_BYTE, REAL) );
}

int
makeImage_Matrix(
	Matrix *matrixP,
	Image **imageP
)
{
	Image *imgP;

	if(!(*imageP))
		if(!(*imageP = (Image *) malloc(sizeof(Image)))) {
			error_CVIP("makeImage_Matrix", "malloc failed!");
			return 0;
		}
	imgP = *imageP;

	if(!(imgP->bandP = (Matrix **) malloc(sizeof(Matrix *)))) {
		error_CVIP("makeImage_Matrix", "malloc failed!");
		return 0;
	}
	*(imgP->bandP) = matrixP;

	imgP->image_format = VIP;
	imgP->color_space = GRAY_SCALE;
	imgP->bands = 1;

return(1);
}

void BinToGray_Image(
	Image *imageP
)
{

	byte *bP;
	int img_size;
	register int i;

	getBandVector_Image(imageP, &bP);

	img_size = getNoOfRows_Image(imageP)*getNoOfCols_Image(imageP);

	for(i=0; i < img_size; i++, bP++)
		if(*bP > 0)
			*bP = 255;
	
}

Image *
cloneNoData_Image(
	Image *inP
)
{
	Image *outP;

	outP = (Image *) malloc(sizeof(Image));
	*outP = *inP;
	outP->bandP = 	(Matrix **) calloc(getNoOfBands_Image(inP),
			sizeof(Matrix *) );
	return(outP);
}


Image *mask_Image(
	Image *inP,
	Image *maskP
)
{
	Image *outP;
	unsigned image_bands, mask_bands;
	register int i, j;

	image_bands = getNoOfBands_Image(inP);
	mask_bands  = getNoOfBands_Image(maskP);

	outP = cloneNoData_Image(inP);

	for(i=0, j=0; i < image_bands; i++) {
		if(j < mask_bands-1) j++;

		outP->bandP[i] = and_Matrix(	getBand_Image(inP, i),
						getBand_Image(maskP, j)
			  	 );
	}

return outP;
}

long 
getArea_Border(
	Border borP
)
{
	byte *bP;
	unsigned img_size;
	long area = 0;

	img_size = getNoOfRows_Image(borP)*getNoOfCols_Image(borP);
	getBandVector_Image(borP, &bP);

	while(img_size--)
		if(*bP++)
			area++;

	return area;
}

Border
xor_Border(
	Border 	bor1P,	/* pointer to first border image */
	Border 	bor2P	/* pointer to first border image */
)
{
	byte		*b1P, *b2P, *bxP, b_mask = 0x01;
	unsigned 	b_size; 	/* size of image border in pixels */
	register int	i;
	Border		borxP;		/* resulting border image */

	b_size = getNoOfRows_Image(bor1P)*getNoOfCols_Image(bor2P);

	/*
	** reference data through band vector
	*/
	getBandVector_Image(bor1P, &b1P);
	getBandVector_Image(bor2P, &b2P);

	/*
	** allocate new border image to hold result of XOR operation
	*/
	borxP = new_Border(getNoOfCols_Image(bor1P), getNoOfRows_Image(bor1P));

	getBandVector_Image(borxP, &bxP);

	for(i=0; i < b_size; i++, b1P++, b2P++, bxP++)
		*bxP = (b_mask & *b1P) ^ (b_mask & *b2P);

	return borxP;
}


double
correlate_Border(
	Border manP,		/* manually segmented border */
	Border autP		/* automatically segmented border */
)
{
	Border 	xorP; 		/* correlation border image */
	double 	correlation; 	/* correlation coefficient */
	long 	man_area, 	/* manually segmented area */
		xor_area;	/* area of nonoverlapping regions */

	man_area = getArea_Border(manP);

	xorP = xor_Border(manP, autP);
	xor_area = getArea_Border(xorP);

	if(man_area == 0) {
		error_CVIP("correlate_Border", "border has zero area");
		return 0;
	}

	correlation = ((double) xor_area)/man_area;

	delete_Image(xorP);	

	return correlation;
}


	
#if 0


double
correlate_Border(
	Border manP,		/* manually segmented border */
	Border autP		/* automatically segmented border */
)
{
	Image 	*corP; 		/* correlation image */
	double 	correlation; 	/* correlation coefficient */
	long 	man_area, 	/* manually segmented area */
		aut_area, 	/* automatically segmented area */
		cor_area;	/* intersection of two areas */

	man_area = getArea_Border(manP);
	aut_area = getArea_Border(autP);

	corP = mask_Image(manP, autP);
	cor_area = getArea_Border(corP);

	if(man_area == 0 || aut_area ==0) {
		error_CVIP("correlate_Border", "image has zero area");
		return 0;
	}

	correlation = cor_area/(sqrt((double)man_area)*sqrt((double)aut_area));

	delete_Image(corP);	

	return correlation;
}

#endif

#if !defined(SKIN)
#define SKIN 1
#endif

Border 
col2Bin_Border(
	Image 	*origP,
	Image 	*cborP
)
{
	ColorHistogram 	*chP; 		/* color map of color border image */
	Border 		skinP,		/* non-skin maskimage */
			bborP;		/* binary border image */
	Color 		colors[2], 	/* colors in color map */
			pixel,		/* current color of interest */
			skin;		/* color most likely to be skin */
	unsigned	rows,		/* number of rows in image */
			cols,		/* number of rows in image */
			imgsize,	/* rows*cols */
			csizes[2] = {0, 0};
	byte 		*bbP,		/* binary border data pointer */
			*sbP,		/* binary mask image data pointer */
			**bvecP,	/* band vector pointer */
			*rP, *gP, *bP;	/* color border data pointers */
	register int	i;
	const char *fn = "col2Bin";

	chP = new_ColorHist();
	if(!chP) {
		error_CVIP(fn, "could not allocate color hist.");
		return NULL;
	}
	compute_ColorHist(chP, cborP, 2);
	if(!chP->histogram) {
		error_CVIP(fn, "too many colors in color border image");
		free(chP);
		return NULL;
	}
	colors[0] = chP->histogram[0].pixel;
	colors[1] = chP->histogram[1].pixel;

	rows = getNoOfRows_Image(cborP);
	cols = getNoOfCols_Image(cborP);
	imgsize = rows*cols;

	bborP = new_Border(cols, rows);
	if(!bborP) {
		error_CVIP(fn, "could not allocate border image");
		delete_ColorHist(chP);
		return NULL;
	}
	getBandVector_Image(bborP, &bbP);

	skinP = findNonSkin_DVis(origP);
	if(!skinP) {
		error_CVIP(fn, "could not find mask image");
		delete_ColorHist(chP);
		delete_Image(bborP);
		return NULL;
	}
	getBandVector_Image(skinP, &sbP);

	bvecP = (byte **) malloc(getNoOfBands_Image(cborP)*sizeof(byte *));
	if(!bvecP) {
		error_CVIP(fn, "could not allocate band vector");
		delete_ColorHist(chP);
		delete_Image(skinP);
		delete_Image(bborP);
		return NULL;
	}
	getBandVector_Image(cborP, bvecP);

	rP = bvecP[RED];
	gP = bvecP[GRN];
	bP = bvecP[BLU];

	for(i=0; i < imgsize; i++, sbP++, rP++, gP++, bP++) {
		assign_Color(pixel, *rP, *gP, *bP);
		if(*sbP)
			if(equal_Color(pixel,colors[0]))
				csizes[0]++;
			else
				csizes[1]++;
	}
	skin = (csizes[0] > csizes[1]) ? colors[0] : colors[1];
		
	rP = bvecP[RED];
	gP = bvecP[GRN];
	bP = bvecP[BLU];

	for(i=0; i < imgsize; i++, bbP++, rP++, gP++, bP++) {
		assign_Color(pixel, *rP, *gP, *bP);
		*bbP = equal_Color(pixel,skin);
	}
		

	free(bvecP);
	delete_ColorHist(chP);
	delete_Image(skinP);

	return bborP;
}

Image *
color2index_Image(
	Image *imageP
)
{
   	Image 		*indexP;
   	Color 		pixel;
   	ColorHashTable 	*chtP;
	ColorHistogram 	*chP;
	unsigned 	rows, cols;
	register int 	i;
	byte 		*iP, *rP, *gP, *bP, **bvecP;

	chP = new_ColorHist();
	compute_ColorHist(chP, imageP, 256);

   	if(!(chP->histogram)) {
      		error_CVIP("color2index", "too many colors!\n");
      		return NULL;
   	}

   	rows = getNoOfRows_Image(imageP);
   	cols = getNoOfCols_Image(imageP);

	bvecP = (byte **) malloc(sizeof(byte *)*3);
	getBandVector_Image(imageP, bvecP);

   	rP = bvecP[RED];
   	gP = bvecP[GRN];
   	bP = bvecP[BLU];

   	indexP = new_Image(	PGM,
				GRAY_SCALE,
				1,
				rows,
				cols,
				CVIP_BYTE, 
				REAL
		 );

   	getBandVector_Image(indexP, &iP);

   	chtP = hist2Hash_ColorHT(chP);

   	for(i=0; i < rows*cols; i++, iP++, rP++, gP++, bP++) {

      		assign_Color(pixel, *rP, *gP, *bP);

      		if( (*iP = lookUpColor_ColorHT( chtP, pixel )) == -1 )
         		return NULL;
   	}

	free(bvecP);
	delete_ColorHist(chP);
   	delete_ColorHT(chtP);

return indexP;
}

#if 0

int
areaCompare(
        const void *o1P,
        const void *o2P
)
{
    return      ((Object *)o2P)->prop.area -
                ((Object *)o1P)->prop.area;
}

#endif

int 
centroidCompare( 
	const void *o1P, 
	const void *o2P
)
{
    return 	( SQR(((Object *)o1P)->prop.h_cog - gmh_cog) +
		  SQR(((Object *)o1P)->prop.v_cog - gmv_cog) )
		-
		( SQR(((Object *)o2P)->prop.h_cog - gmh_cog) +
		  SQR(((Object *)o2P)->prop.v_cog - gmv_cog) );
}


int
getBest_Objects(
	ObjectList 	olP
)
{
	ObjectVector 	*ovP;
	Object		*oP;
	int 		label_best;

	ovP = listToVector_Objects(olP);

	if(!ovP) {
		error_CVIP("getBest", "could not allocate object vector");
		return 0;
	} 

        qsort(ovP->oP, ovP->no_of_objects, sizeof(Object), &(centroidCompare));

	label_best = ovP->oP[0].label;

	head_LL(olP);
	previous_LL(olP);

	for (;;) {
        	oP = retrieveNext_LL(olP);

		if(oP->label != label_best)
			removenext_LL(olP);
		else
			next_LL(olP);

		if (istail_LL(olP)) break;
	}

	free(ovP->oP);
	free(ovP);

return 1;
}



int
getLocDat_Image(
	Image *imgP,
	int value,
	int *ray_x, 
	int *ray_y
)
{
	int		*iP;
	register int	y, x;
	unsigned	rows, cols;

	rows = getNoOfRows_Image(imgP);
	cols = getNoOfCols_Image(imgP);
	getBandVector_Image(imgP, &iP);

	for(y=0; y < rows; y++)
		for(x=0; x < cols; x++, iP++)
			if(*iP == value) {
				*ray_y = y;
				*ray_x = x;
				return 1;
			}

	return 0;
}

#if 0
CVIP_BOOLEAN
lookUp(
	Image	*imgP,
	int	x,
	int	y
)
{
	register int	i;

	for(i=y; i >= 0;
#endif

int
getInsidePoint_Contour(
	Image *imgP,
	int value,
	int *ray_x, 
	int *ray_y
)
{
	byte		*bP;
	register int	y, x;
	unsigned	rows, cols, point1, point2;
	byte		flag;
	CVIP_BOOLEAN		firstpoint = CVIP_YES;

	rows = getNoOfRows_Image(imgP);
	cols = getNoOfCols_Image(imgP);

	for(y=0; y < rows; y++) {

		bP = getRow_Image(imgP, y, 0);
		for(x=0, flag = 0; x < cols; x++, bP++)
			if(*bP) {
				if(firstpoint == CVIP_YES) {
					firstpoint = CVIP_NO;
					break;
				}
			
				flag = ~flag;

				if(flag) {
					while( (*bP) && (x < cols) ) {
						bP++;
						x++;
					}
					bP--;
					x--;
					point1 = x;
				}
				else {
					point2 = x;
					while( (*bP) && (x < cols) ) {
						bP++;
						x++;
					}
					bP--;
					x--;
					if( (point2 - point1) > 2 ) {
						*ray_y = y;
						*ray_x = point1+(point2 - point1)/2;
						return 1;
					}
				}
			}
	}

	return 0;
}


int
PMedian_Image(
	Image 	*imgP
)
{
	Matrix		*bufP, *datP;
	byte 		**biP, **bmP, *y1P, *y2P, *xiP, *xmP;
	byte		x1, xm, xc, y1, yr, mminx, mminy, mmaxx, mmaxy;
	register int	i, j, k;
	unsigned 	bands, rows, cols;

	cols	= getNoOfCols_Image(imgP);
	rows	= getNoOfRows_Image(imgP);
	bands 	= getNoOfBands_Image(imgP);

	if(!(bufP = new_Matrix(rows, cols, CVIP_BYTE, REAL))) {
		error_CVIP("PMedian", "could not allocate matrix!");
		return 0;
	}

	for(k=0; k < bands; k++) {
		datP = getBand_Image(imgP, k);
		(void) copy_Matrix(datP, bufP);
		biP = getData_Matrix(datP);
		bmP = getData_Matrix(bufP);
		for(i=0; i < rows-2; i++) {
			y1P = biP[i]+1;
			y2P = biP[i+2]+1;
			xiP = biP[i+1];
			xmP = bmP[i+1]+1;
			x1 = *xiP++;
			xm = *xiP++;
			xc = *xiP++;
			y1 = *y1P++;
			yr = *y2P++;
			for(j=2; j < cols; j++, y1P++, y2P++, xiP++, xmP++) {
				mminx = MAX((byte)MIN(x1,xm),(byte)MIN(xm,xc));
				mminy = MAX((byte)MIN(y1,xm),(byte)MIN(xm,yr));
				mmaxx = MIN((byte)MAX(x1,xm),(byte)MAX(xm,xc));
				mmaxy = MIN((byte)MAX(y1,xm),(byte)MAX(xm,yr));

				*xmP = 	((byte) MAX(mminx, mminy) >> 1) +
				      	((byte) MIN(mmaxx, mmaxy) >> 1);

				x1 = xm;
				xm = xc;
				xc = *xiP;
				y1 = *y1P;
				yr = *y2P;
			}
		}
		(void) copy_Matrix(bufP, datP);	
	}

	delete_Matrix(bufP);

	return 1;
}
				

int
findNoColors_Image(
	Image	*imgP
)
{
	Color		bgnd = {0, 0, 0};
	byte 		**bvecP, max[3] = {0, 0, 0}, 
			min[3] = {255, 255, 255};
	float 		var[3], ave[3], sum[3] = {0, 0, 0}, 
			sumsqr[3] = {0, 0, 0};
	register int	i, j;
	unsigned	count = (0), rows, cols, imgsize;

	bvecP = (byte **) malloc(3*sizeof(byte *));
	getBandVector_Image(imgP, bvecP);

	rows = getNoOfRows_Image(imgP);
	cols = getNoOfCols_Image(imgP);
	imgsize = rows*cols;

	for(i=0; i < imgsize; i++) {
			
			if(!(	bvecP[RED][i] == getRed_Color(bgnd) &&
				bvecP[GRN][i] == getGrn_Color(bgnd) &&
				bvecP[BLU][i] == getBlu_Color(bgnd)
			)) {
				count++;
				for(j=0; j < 3; j++) {
					max[j] = MAX(bvecP[j][i], max[j]);
					min[j] = MIN(bvecP[j][i], min[j]);
           				sum[j] += bvecP[j][i];
           				sumsqr[j] += (float) bvecP[j][i]*bvecP[j][i];
				}
			}

	}

	for (j=0; j < 3; j++) {

        	ave[j] = sum[j]/count;
        	var[j] = sumsqr[j]/count - ave[j]*ave[j];
		fprintf(stderr, "max[%d] = %d.\n", j, max[j]);
		fprintf(stderr, "min[%d] = %d.\n", j, min[j]);
		fprintf(stderr, "ave[%d] = %.4f.\n", j, ave[j]);
		fprintf(stderr, "var[%d] = %.4f.\n", j, var[j]);

       	}

	maxP = (byte *) max;
	minP = (byte *) min;
	aveP = (float *) ave;
	varP = (float *) var;
	/*
	fprintf(stderr, "\nnumber of colors is %d\n", fnumcol6());
	*/
}

float 
findnum(
	const char *selP
)
{
	if(	!strcmp(selP, "RMIN") )
		return minP[RED];
	else if(!strcmp(selP, "GMIN") )
		return minP[GRN];
	else if(!strcmp(selP, "BMIN") )
		return minP[BLU];
	else if(!strcmp(selP, "RMAX"))
		return maxP[RED];
	else if(!strcmp(selP, "GMAX"))
		return maxP[GRN];
	else if(!strcmp(selP, "BMAX"))
		return maxP[BLU];
	else if(!strcmp(selP, "RAVE"))
		return aveP[RED];
	else if(!strcmp(selP, "GAVE"))
		return aveP[GRN];
	else if(!strcmp(selP, "BAVE"))
		return aveP[BLU];
	else if(!strcmp(selP, "RVAR"))
		return varP[RED];
	else if(!strcmp(selP, "GVAR"))
		return varP[GRN];
	else if(!strcmp(selP, "BVAR"))
		return varP[BLU];
	else {
		error_CVIP("findnum", "invalid 1st class variable name");
		return -1.0;
	}
}

	

CVIP_BOOLEAN 
stdDevTest(
	Image 	*imgP, 
	QUAD 	*quadP, 
	void 	*paramP
)
{

   	register int 	i, j;
   	unsigned int 	yoffset, meansum = (0), varsum = (0),
			no_of_pixels, count_pixels = (0);
	float		var, stddev, mean;
   	byte 		*bP, *meanP = (byte *) paramP, bdata;
   	char 		*fn = "mean_test";

   	yoffset = quadP->y + quadP->dy;

      	for(j = quadP->y; j < yoffset; j++) {
         	bP  = (byte *) imgP->bandP[0]->rptr[j] + quadP->x;
            	for(i= 0; i < quadP->dx; i++, bP++) {
			if(*bP) {
				count_pixels++;
               			meansum += *bP;
				varsum	+= (*bP)*(*bP);
			}
		}
     	}

	if(!count_pixels)
		count_pixels = 1;

      	mean 	= ((float)meansum)/count_pixels;
 	var 	= ((float)varsum)/count_pixels - mean*mean;
   	stddev 	= (float) sqrt((double)var);

   	if(stddev < 10.0) {
		bdata = (byte) ROUND(mean);
		if(bdata >= *meanP)
      			quadP->data[0] = 255;
		else
			quadP->data[0] = 1;
      		return CVIP_YES;
   	}
   	else if (quadP->dx==1 || quadP->dy==1) {
      		quadP->data[0] = 0;
      		return CVIP_YES;
   	}
   	else 
      		return CVIP_NO;
}
			 

ObjectList
dupList_Objects(
	ObjectList	listP
)
{
	ObjectList	newP;
	Object		*objP;
	slink 		*refP = listP->clP;

	newP = new_LL();

	head_LL(listP);

   	if( !isempty_LL(listP) ) {
      		for (;;) {
			objP = (Object *) malloc(sizeof(Object));
			(void) memcpy((void *) objP, (void *)retrieve_LL(listP),
			sizeof(Object));
			addnext_LL(newP, objP);
         		if (istail_LL(listP)) break;
         			next_LL(listP);
      		}
   	}

   	listP->clP = refP;

	return newP;
}

int 
numberCompare( 
	const void *o1P, 
	const void *o2P
)
{
    return 	((CentroidObject) o1P)->number - 
		((CentroidObject) o2P)->number;
}

CentroidObject
findCentroidObject(
	CentroidObject  objP,
	int		n,
	int		number
)
{
	CentroidObjectType	obj;

	obj.number = number;

	return 
	bsearch(&obj, objP, n, sizeof(CentroidObjectType), &(numberCompare));
}

int
readCentroidFile(
	const char 	*fname,
	CentroidObject	*objP,
	int		n
)
{
	FILE 		*fileP;
	const char	*fn = "readCentroidFile";
	char		buf[80];
	register int	i;
	int		status;

	*objP = (CentroidObject) malloc(n*sizeof(CentroidObjectType));

	if(!(fileP = openRead_CVIP(fname))) {
		error_CVIP(fn, "could not open centroid file");
		return 0;
	}

	for(i=0; i < n; i++) {

		if(fGetLine_CVIP(fileP, buf) != 1) {
			error_CVIP(fn, "%d tumor entries not found", n);
			break;
		}

		(*objP)[i].number = 
		(int) strGetInt_CVIP(buf, &status, 1, ":\t");
		if(!status) {
			error_CVIP(fn, "bad tumor number entry");
			break;
		}

		(*objP)[i].h_cog = 
		(float) strGetFloat_CVIP(buf, &status, 2, ":\t");
		if(!status) {
			error_CVIP(fn, "bad tumor h_cog entry");
			break;
		}


		(*objP)[i].v_cog = 
		(float) strGetFloat_CVIP(buf, &status, 3, ":\t");
		if(!status) {
			error_CVIP(fn, "bad tumor v_cog entry");
			break;
		}

	}

	if(!status || (i != n)) {
		free(*objP);
		close_CVIP(fileP);
		return(0);
	}

 	qsort(*objP, n, sizeof(CentroidObjectType), &(numberCompare));

	close_CVIP(fileP);
	return 1;
}

int
drawTumorBorder_Image(
	Image	*imgP,
	Border	borP,
	byte	*color
)
{
	byte		**bvecP, *bborP;
	register int	i, j;
	unsigned	bands, imgsize;

	bands = getNoOfBands_Image(imgP);
	imgsize = getNoOfRows_Image(imgP)*getNoOfCols_Image(imgP);

	getBandVector_Image(borP, &bborP);

	bvecP = (byte **) malloc(sizeof(byte *)*bands);
	getBandVector_Image(imgP, bvecP);

	for(i=0; i < imgsize; i++, bborP++)
		if(*bborP & 0x01)
			for(j=0; j < bands; j++)
				bvecP[j][i] = color[j];

	free(bvecP);
	return 1;
}

int
vectorMoment(
	float		*data,
	int		n,
	float		*ave,
	float		*sdev
)
{
	register int	j;
	float		s, svar;

	*ave 	= 0.0;
	*sdev 	= 0.0;
 	s	= 0.0;

        for (j=0;j<n;j++)
		s += data[j];
        *ave=s/n;

        svar=0.0;
        for (j=0;j<n;j++)
                svar += SQR(data[j]-(*ave));
               
        svar /= (n-1);
        *sdev=sqrt(svar);
}
		



		

		
	

