/*
 * Image Processing / Computer Vision Tool Project - Dr. Umbaugh, SIUE
 * restore_xformfilter.c
 * 
 * Description: Contains routines necessary to perform
 * an inverse filter.  The inverse filter is a Frequency Domain
 * Filter defined as follows:
 *
 *          G(u,v)
 * F(u,v) = ------
 *          H(u,v)
 *
 * Where F(u,v) is the Fourier Transform of the restored image;
 *       G(u,v) is the Fourier Transform of the degraded image; and
 *       H(u,v) is the Fourier Transform of the degradation function.
 *
 * Credits: Arve Kjoelen
 */
#include "CVIPimage.h"
#include "CVIPtransform.h"
#include "CVIPxformfilter.h"
#include "CVIPdef.h"
#include "CVIParithlogic.h"

/*
 * Routine Name: restor_filter_Setup()
 *
 * Description:
 *	Setup function for all restoration routines.  An equation for these
 *	routines, parameterized in terms of gamma and alpha, is:
 *
 *              _       _ alpha     _                                _ (1-alpha)
 *             | H*(u,v) |         |                  H*(u,v)         |
 * FIL=G(u,v)* | ------- |         | -------------------------------- |
 *             |        2|         |        2                         |
 *             ||H(u,v)| |         ||H(u,v)| + gamma(Pn(u,v)/Pf(u,v)) |
 *              -       -           -                                -
 *
 * Variables passed: 
 *	Image *degraded - pointer to degraded image ( H(u,v) ).
 *	Image *degradation_fn - pointer to degradation function. Ideally, 
 *	we want the inverse of this (G(u,v) )
 *	Image *noise_fn - pointer to the noise function. (Pn(u,v) is the power 
 *	spectrum of this.
 *	float gamma - gamma in the above equation restr. to the interval [0,1].
 *	float alpha - alpha in the above equation. Restr. to the interval [0,1].
 *	int filter - filter to be performed:
 *		0 = Inverse Filter	(obtained by alpha=0 and gamma=0 above)
 *		1 = Simple Wiener Filter (obtained by alpha=0 and substituting
 *		    a constant, k, for [gamma(Pn(u,v)/Pf(u,v)]
 *		    above.  This constant is passed as gamma).
 *		2 = Wiener Filter (obtained by alpha=0 and gamma=1)
 *		3 = Parametric Wiener Filter    (obtained by alpha=0 above).
 *		4 = Power Spectrum Equalization (alpha=0.5 and gamma=1)
 *		5 = Least-squares restoration   (obtained by letting alpha=0 and
 *		    replacing Pn(u,v)/Pf(u,v) with the Fourier Transform
 *	 	    of the highpass mask
 *
 *							 0	-1	 0
 *
 *							-1	 4	-1
 *
 *							 0	-1	 0
 *
 *		    this mask is zero-padded to the size of
 *		    the degraded image.
 *		6 = Geometric Mean Filter (the parameterized equation above)
 *		
 *	int choice - Specifies how to handle denominator=0 
 *	int cutoff - specifies a cutoff frequency for filtering
 *
 *	NOTE: The above images may be passed as either spatial-domain images or
 *	      transformed (w/FFT) images.  history_check() is used to determine
 *	      whether the images are in the transform domain.  the FFT is per-
 *	      formed as needed if the passed images are in the spatial domain.
 *
 */

/*
 * function declarations
 */
static int inverseFilter_Image(Image **, Image *, Image *, float, int);
static int idealFilter_Matrix(Matrix **, float, int, int);
static int block_power_2(int);
static Image *pse_n_gm_filter(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, int choice, int cutoff, int root);
/*
 * Routine Name: simple_wiener()
 *
 * Description: 
 * 	Performs the operations needed to perform a simple Wiener Filter
 *	Denoting the degraded image by G(u,v), and using
 *	H(u,v) for the degradation function Image, this function calculates
 *		                    2
 *                          |H(u,v)|
 *        G(u,v)    *    --------------
 *                               2
 *                       |H(u,v)|   + k
 *
 * Variables Passed:
 *	Image *degr - Contains the degraded image.
 *	Image *degr_fn - Contains the degradation function.
 *	float k - constant.
 *
 * Diagnostics:  
 *    	- This function does its own memory management.  if you'd
 *    	like to keep you images intact, make a copy of them before you pass
 * 	them to this function.
 *	- The function tries to determine whether an FFT has been performed
 *	on the passed images.  If not, the function will perform the FFT.
 *	- the function will also extend the sizes of the passed images as best
 *	it can when they are of unequal size.
 *
 * Author: Arve Kjoelen, SIUE
 *
 */
Image *simple_wiener(Image *degr, Image *degr_fn, float k) {
   unsigned cnt, i,j;
   unsigned rows, cols;
   CVIP_TYPE tp1,tp2;
   Image *copy;
   HISTORY stuff=NULL;
   int block_size=0, *hptr;
   float *fptr;
   Matrix *temp_Matrix=NULL;

   rows=degr->image_ptr[0]->rows;
   cols=degr->image_ptr[0]->cols;
   tp1=degr->image_ptr[0]->data_type;
   tp2=degr_fn->image_ptr[0]->data_type;

/*
 * First order of business:  Check whether images have had the FFT performed.
 * if not, do so and add history entry for FFT.
 */
   block_size=MAX(degr->bandP[0]->rows,degr->bandP[0]->cols);
   block_size=MAX(block_size,MAX(degr_fn->bandP[0]->rows,degr_fn->bandP[0]->cols));
   /* Function to convert the block size to a power of 2 */
   block_size=block_power_2(block_size);

   if(!history_check(FFT, degr)) {
       fprintf(stdout,  "The degraded image appears to be in the spatial domain\n");
       fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
       degr=fft_transform(degr, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(degr, stuff);
       stuff=NULL;
   }

   if(!history_check(FFT, degr_fn)) {
       fprintf(stdout, "The degradation function appears to be in the spatial domain\n");
       fprintf(stdout, "Performing FFT and any necessary zero-padding...\n");
       degr_fn=fft_transform(degr_fn, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(degr_fn, stuff);
       stuff=NULL;
   }


   if(tp1<CVIP_FLOAT) { 
      if(tp2!=CVIP_DOUBLE)
         (void)cast_Image(degr, CVIP_FLOAT);
      else
         (void)cast_Image(degr, CVIP_DOUBLE);
   }
   if(tp2<CVIP_FLOAT) {
      if(tp1!=CVIP_DOUBLE)
         (void)cast_Image(degr_fn, CVIP_FLOAT);
      else
         (void)cast_Image(degr_fn, CVIP_DOUBLE);
   }
   tp1=degr->image_ptr[0]->data_type;

/* Make a copy of the degradation function and calculate
 * its conjugate.
 */
   fprintf(stdout, "Performing simple Wiener Filter...\n");
   copy=duplicate_Image((const Image *)degr_fn);
   for(i=0; i<copy->bands; i++) {
      if(conj_Matrix(copy->bandP[i])==CVIP_NO) {
	 fprintf(stderr, "Error in conj_Matrix.  Bailing...\n");
         exit(1);
      }
   }

   
   for(cnt=0; cnt<degr->bands; cnt++) {
      temp_Matrix=degr_fn->image_ptr[cnt];
      degr_fn->image_ptr[cnt]=square_mag_Matrix((const Matrix *)degr_fn->image_ptr[cnt]);
      delete_Matrix(temp_Matrix);
/* Check for data format because trying to access imaginary part of a real image will result in a segmentation fault */
      for(i=0; i<rows; i++) {
            for(j=0; j<cols; j++) {
    		if(tp1==CVIP_FLOAT) {
               /*((float *)copy->image_ptr[cnt]->rptr[i])[j]/=(((float *)degr_fn->image_ptr[cnt]->rptr[i])[j]+k);
               ((float *)copy->image_ptr[cnt]->iptr[i])[j]/=(((float *)degr_fn->image_ptr[cnt]->rptr[i])[j]+k);*/
		((float **)(copy->image_ptr[cnt]->rptr))[i][j]/=(((float **)(degr_fn->image_ptr[cnt]->rptr))[i][j]+k);
		if (degr_fn->image_ptr[0]->data_format==COMPLEX)
			((float **)(copy->image_ptr[cnt]->iptr))[i][j]/=(((float **)(degr_fn->image_ptr[cnt]->iptr))[i][j]+k);
	        }
        	 else { /* CVIP_DOUBLE */
	       /*((double *)copy->image_ptr[cnt]->rptr[i])[j]/=(((double *)degr_fn->image_ptr[cnt]->rptr[i])[j]+k);
	       ((double *)copy->image_ptr[cnt]->iptr[i])[j]/=(((double *)degr_fn->image_ptr[cnt]->rptr[i])[j]+k);*/
		((double **)(copy->image_ptr[cnt]->rptr))[i][j]/=(((double **)(degr_fn->image_ptr[cnt]->rptr))[i][j]+k);
		if (degr_fn->image_ptr[0]->data_format==COMPLEX)
			((double **)(copy->image_ptr[cnt]->iptr))[i][j]/=(((double **)(degr_fn->image_ptr[cnt]->iptr))[i][j]+k);
	        }
	    }
      }
      temp_Matrix=copy->image_ptr[cnt]; 
      copy->image_ptr[cnt]=multPWise_Matrix((const Matrix *)degr->image_ptr[cnt], (const Matrix *)copy->image_ptr[cnt]);
      /* The temporary matrix is deleted because it is not deleted by the multPWise_Matrix function */
      delete_Matrix(temp_Matrix);
   }
/*
 * Record history information
 */
   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=1;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_FLOAT;
   fptr=(float *)malloc(sizeof(float));
   fptr[0]=k;
   stuff->packetP->dptr = (void**)malloc(sizeof(void*));
   stuff->packetP->dptr[0]=(void *)fptr;
   stuff->packetP->dsize = 1;
   stuff->next=(HISTORY)NULL;
   stuff->ftag=SIMPLE_WIENER;
   history_add(copy, stuff);
   stuff=NULL;
   delete_Image(degr_fn);
   delete_Image(degr);
   return(copy);
}

/*
 * Routine name least_squares()
 *
 * Description: see header.  The least_squares_filter() routine must, because
 * of its potential for zero in the denominator, give the user the option
 * of performing some additional hanky-panky when division blows up.
 * The function takes as its arguments:
 *    Image *degr 	- pointer to the degraded image.
 *    Image *degr_fn 	- pointer to the degradation function.
 *    Image *snr_approx - pointer to image that approximates Pn(u,v)/Pf(u,v).
 *                        In many practical applications, this is the FFT
 *			  of the highpass mask specified at the top of this
 *			  file.  However, any image may be passed.
 *    float gamma	- gamma in equation above.
 *
 * Diagnostics:  
 *    	- This function does its own memory management.  if you'd
 *    	like to keep you images intact, make a copy of them before you pass
 * 	them to this function.
 *	- The function tries to determine whether an FFT has been performed
 *	on the passed images.  If not, the function will perform the FFT.
 *	- the function will also extend the sizes of the passed images as best
 *	it can when they are of unequal size.
 *
 * Returns: Pointer to the restored image (in the frequency domain).
 * 
 * Author: Arve Kjoelen, SIUE
 */
Image *least_squares(Image *degr, Image *degr_fn, Image *snr_approx, float gamma, int choice, int cutoff) {
   Image *sq_mag, *retImg;
   int i, block_size, *hptr;
   HISTORY stuff;
   CVIP_TYPE tp1, tp2;
   Matrix *temp_Matrix=NULL;
   Image *temp_Image=NULL;

   tp1=degr->image_ptr[0]->data_type;
   tp2=degr_fn->image_ptr[0]->data_type;

/*
 * First order of business:  Check whether images have had the FFT performed.
 * if not, do so and add history entry for FFT.
 */
   block_size=MAX(degr->bandP[0]->rows,degr->bandP[0]->cols);
   block_size=MAX(block_size,MAX(degr_fn->bandP[0]->rows,degr_fn->bandP[0]->cols));
   /* Function to convert the block size to a power of 2 */
    block_size=block_power_2(block_size);

   if(!history_check(FFT, degr)) {
       fprintf(stdout,  "The degraded image appears to be in the spatial domain\n");
       fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
       degr=fft_transform(degr, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(degr, stuff);
       stuff=NULL;
   }

   if(!history_check(FFT, degr_fn)) {
       fprintf(stdout, "The degradation function appears to be in the spatial domain\n");
       fprintf(stdout, "Performing FFT and any necessary zero-padding...\n");
       degr_fn=fft_transform(degr_fn, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(degr_fn, stuff);
       stuff=NULL;
   }

if(!history_check(FFT, snr_approx)) {
       fprintf(stdout, "The SNR approximation appears to be in the spatial domain\n");
       fprintf(stdout, "Performing FFT and any necessary zero-padding...\n");
       snr_approx=fft_transform(snr_approx, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(snr_approx, stuff);
       stuff=NULL;
   }


   tp1=degr->image_ptr[0]->data_type;

   if(tp1<CVIP_FLOAT) {
      if(tp2!=CVIP_DOUBLE)
         (void)cast_Image(degr, CVIP_FLOAT);
      else
         (void)cast_Image(degr, CVIP_DOUBLE);
   }
   if(tp2<CVIP_FLOAT) {
      if(tp1!=CVIP_DOUBLE)
         (void)cast_Image(degr_fn, CVIP_FLOAT);
      else
         (void)cast_Image(degr_fn, CVIP_DOUBLE);
   }

   if(snr_approx->image_ptr[0]->data_type != degr->image_ptr[0]->data_type)
      (void)cast_Image(snr_approx, degr->image_ptr[0]->data_type);

   fprintf(stdout, "Performing Least-squares Restoration Filter...\n");
/*
 * Scale everything.
 * Divide by N, since our definition of the FFT
 * uses 1/N instead of 1/(N*N) as the forward multiplier.
 * While we're at it, make sure snr_approx gets multiplied by gamma.
 */
   for(i=0; i<degr_fn->bands; i++) {
      degr->bandP[i]=scale_Matrix(degr->bandP[i], 1.0/(double)degr->bandP[i]->cols);
      degr_fn->bandP[i]=scale_Matrix(degr_fn->bandP[i], 1.0/(double)degr_fn->bandP[i]->cols);
      snr_approx->bandP[i]=scale_Matrix(snr_approx->bandP[i], (double)gamma/(double)snr_approx->bandP[i]->cols);
   }


/*
 * We're now done with setup, checking, type casting
 * and so on.  Start producing some of the elements of
 * the constrained least squares equation.
 */
   sq_mag=new_Image(degr_fn->image_format, degr_fn->color_space, degr_fn->bands, degr_fn->bandP[0]->rows, degr_fn->bandP[0]->cols, degr_fn->bandP[0]->data_type, degr_fn->bandP[0]->data_format);
   /*sq_mag=duplicate_Image((const Image *)degr_fn);*/
   for(i=0; i<degr_fn->bands; i++) {
      sq_mag->image_ptr[i]=square_mag_Matrix((const Matrix *)degr_fn->image_ptr[i]);
   }
   /*temp_Image=sq_mag;*/

   sq_mag=add_Image(sq_mag, snr_approx);
   /*delete_Image(temp_Image);*/
   /*delete_Image(snr_approx);*/

/*
 * At this point, we'll call the inverse filter core routine.  All this
 * does is divide degr(G[u,v]) by (sq_mag + gamma*snr_approx)
 */
   (void)inverseFilter_Image(&retImg, degr, sq_mag, (float)cutoff, choice);
   delete_Image(sq_mag);
   delete_Image(degr);

/*
 * Now, all that's left to complete the calculation of our equation,
 * is to multiply by H*[u,v]
 */
   for(i=0; i<retImg->bands; i++) {
      if(conj_Matrix(degr_fn->bandP[i])==CVIP_NO) {
         fprintf(stderr, "Error in conj_Matrix.  Bailing...\n");
         exit(1);
      }
      temp_Matrix=retImg->bandP[i];
      retImg->bandP[i]=multPWise_Matrix((const Matrix *)retImg->bandP[i], (const Matrix *)degr_fn->bandP[i]);
      delete_Matrix(temp_Matrix);

/*
 * scale back our return matrix, free memory and we're done.
 */
      retImg->bandP[i]=scale_Matrix(retImg->bandP[i], (double)retImg->bandP[i]->cols);
   }
   delete_Image(degr_fn);
/*
 * Before returning, add history
 * information to the image.
 */
   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=1;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_INTEGER;
   hptr=(int *)malloc(sizeof(int));
   hptr[0]=block_size;
   stuff->packetP->dptr = (void**)malloc(sizeof(void*));
   stuff->packetP->dptr[0]=(void *)hptr;
   stuff->packetP->dsize = 1;
   stuff->next=(HISTORY)NULL;
   stuff->ftag=FFT;
   history_add(retImg, stuff);
   stuff=NULL;

   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=2;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_INTEGER;
   stuff->packetP->dtype[1]=CVIP_INTEGER;
   stuff->packetP->dsize = 2;
   stuff->packetP->dptr = (void**)malloc(sizeof(void*)*stuff->packetP->dsize);
   stuff->packetP->dptr[0]=(void *)malloc(sizeof(int));
   stuff->packetP->dptr[1]=(void *)malloc(sizeof(int));
   memcpy((void *)stuff->packetP->dptr[0], (const void *)&choice, sizeof(int));
   memcpy((void *)stuff->packetP->dptr[1], (const void *)&cutoff, sizeof(int));
   stuff->next=(HISTORY)NULL;
   stuff->ftag=LEAST_SQUARES;
   history_add(retImg, stuff);
   stuff=NULL;
   return(retImg);   
   
}

/*
 * Routine Name: wiener()
 *
 * Description:
 *	The Wiener Filter is described in the header of this file.
 *	This function implements a wiener filter from the parameters
 * 	passed to it (G,H,Pn,Pf).  This routine may be used to implement
 *	a parametric wiener filter by multiplying Pn by gamma before
 *	calling this function (again, see equation at top of file).
 * Diagnostics:
 *    	- This function does its own memory management.  if you'd
 *    	like to keep you images intact, make a copy of them before you pass
 * 	them to this function.
 *	- The function tries to determine whether an FFT has been performed
 *	on the passed images.  If not, the function will perform the FFT.
 *	- The function will also extend the sizes of the passed images as best
 *	it can when they are of unequal size.
 *      - The input images are assumed to be spatial-domain if no
 *	FFT is detected in the image history.  This is especially
 *	important to note for the power images.
 *	See also comments in the code for other diagnostics.
 *
 * Author: Arve Kjoelen
 *
 * Date: 11/01/96
 *
 */
Image *wiener(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, int choice, int cutoff) {
   CVIP_TYPE dtype=CVIP_BYTE;	/* data type of incoming images: byte, float, etc. */
   HISTORY stuff=NULL;
   int rmax=0, cmax=0, *hptr=NULL, i;
   Image *retImg=NULL, *conj_Img=NULL;
   Matrix *tpMat=NULL;
   Image *temp_Image=NULL;

   if((degr->bands!=degr_fn->bands) || (degr->bands!=p_noise->bands) || (degr->bands!=p_orig->bands)) {
      fprintf(stderr, "wiener: error:  All images must have the same number of bands\n");
      delete_Image(degr_fn);
      delete_Image(p_noise);
      delete_Image(p_orig);
      return(degr);
   }
/*
 * Find out whether the FFT has been performed on all images.
 * If not, find the largest dimensions present, and perform the
 * FFT on all images to be that size.
 */
   if(!history_check(FFT,degr) || !history_check(FFT, degr_fn) || !history_check(FFT, p_noise) || !history_check(FFT, p_orig)) {
      rmax=MAX(degr->bandP[0]->rows,degr_fn->bandP[0]->rows);
      cmax=MAX(degr->bandP[0]->cols,degr_fn->bandP[0]->cols);
      rmax=MAX(rmax,p_noise->bandP[0]->rows);
      cmax=MAX(cmax,p_noise->bandP[0]->cols);
      rmax=MAX(rmax,p_orig->bandP[0]->rows);
      cmax=MAX(cmax,p_orig->bandP[0]->cols);
      rmax=MAX(rmax,cmax);
      /* Function to convert the block size to a power of 2 */
      rmax=block_power_2(rmax);

      if(!history_check(FFT, degr)) {
         fprintf(stdout,  "The degraded image appears to be in the spatial domain\n");
         fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
         degr=fft_transform(degr, rmax);
      }

      if(!history_check(FFT, degr_fn)) {
         fprintf(stdout,  "The degradation function appears to be in the spatial domain\n");
         fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
         degr_fn=fft_transform(degr_fn, rmax);
      }
/*
 * for p_orig and p_noise,
 * square the result of the FFT.
 */

      if(!history_check(FFT, p_noise)) {
         fprintf(stdout,  "The noise power image appears to be in the spatial domain\n");
         fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
         p_noise=fft_transform(p_noise, rmax);
         for(i=0; i<p_noise->bands; i++) {
	    tpMat=p_noise->image_ptr[i];
	    p_noise->image_ptr[i]=square_mag_Matrix((const Matrix *)p_noise->image_ptr[i]);
	    delete_Matrix(tpMat);
         }
      }

      if(!history_check(FFT, p_orig)) {
         fprintf(stdout,  "The original power image appears to be in the spatial domain\n");
         fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
         p_orig=fft_transform(p_orig, rmax);
         for(i=0; i<p_orig->bands; i++) {
	    tpMat=p_orig->image_ptr[i];
	    p_orig->image_ptr[i]=square_mag_Matrix((const Matrix *)p_orig->image_ptr[i]);
	    delete_Matrix(tpMat);
	 }
      }
   }

/*
 * Next, cast all images to the most
 * precise data type present.
 */
   dtype=degr->image_ptr[0]->data_type;
   dtype=MAX(dtype,degr_fn->image_ptr[0]->data_type);
   dtype=MAX(dtype,p_noise->image_ptr[0]->data_type);
   dtype=MAX(dtype,p_orig->image_ptr[0]->data_type);

   if(degr->image_ptr[0]->data_type<dtype)
      (void)cast_Image(degr,dtype);
   if(degr_fn->image_ptr[0]->data_type<dtype)
      (void)cast_Image(degr_fn,dtype);
   if(p_noise->image_ptr[0]->data_type<dtype)
      (void)cast_Image(p_noise,dtype);
   if(p_orig->image_ptr[0]->data_type<dtype)
      (void)cast_Image(p_orig,dtype);

/*
 * We now have to divide everything by N, since our implementation
 * of the FFT divides by N going forward and inverse, instead of
 * the mathematical definition, which divides by N squared going
 * forward, and 1 going backwards.  However, there's no reason to be
 * pedantic about this.  p_noise is divided by p_orig, so this ratio
 * won't change by division.  Therefore, we only divide degr and degr_fn.
 */
   for(i=0; i<degr->bands; i++) {
      degr->bandP[i]=scale_Matrix(degr->bandP[i], (float)(1.0/rmax));
      degr_fn->bandP[i]=scale_Matrix(degr_fn->bandP[i], (float)(1.0/rmax));
   }

/*
 * We're finally ready to proceed.  First, call the 
 * Inverse Filter routine.  This merely divides
 * p_noise by p_orig, with special processing
 * for when our denominator is 0.
 */
   (void)inverseFilter_Image(&retImg, p_noise, p_orig, (float)cutoff,choice);
   delete_Image(p_noise);
/*
 * Next, we want to add H(u,v) squared to complete our
 * denominator, but it would be best to calculate
 * H*(u,v) first, since we'll need this factor in a bit.
 */
   conj_Img=duplicate_Image(degr_fn);
   for(i=0; i<degr_fn->bands; i++) {
      if(conj_Matrix(conj_Img->bandP[i])==CVIP_NO) {
         fprintf(stderr, "wiener: Error in conj_Matrix.  Bailing...\n");
         exit(1);
      }
      tpMat=degr_fn->image_ptr[i];
      degr_fn->bandP[i]=square_mag_Matrix((const Matrix *)degr_fn->bandP[i]);
      delete_Matrix(tpMat);
   }
/* * Add H(u,v) squared to p_noise/p_orig.  * This completes our denominator.  */
   /*temp_Image=degr_fn;*/
   degr_fn=add_Image(degr_fn,retImg);  
   /*delete_Image(temp_Image);*/
   /*delete_Image(retImg);*/ /* Needed because inverseFilter_Image()
			    will reallocate this when called again */

/*
 * There is a possibility that the entire denominator is 0.
 * Therefore, we will again call inverseFilter_Image() (instead of doing
 * regular division) to take care of those special cases.
 * 
 */
   (void)inverseFilter_Image(&retImg, conj_Img, degr_fn, (float)cutoff,choice);
   delete_Image(degr_fn);
   delete_Image(conj_Img);

/* 
 * Finally, we will multiply by G(u,v).
 */
   for(i=0; i<degr->bands; i++) {
      tpMat=retImg->bandP[i];
      retImg->bandP[i]=multPWise_Matrix((const Matrix *)degr->bandP[i], (const Matrix *)retImg->bandP[i]);
      delete_Matrix(tpMat);
   }
/*
 * We are done using degr
 */
   delete_Image(degr);

/*
 * We have decided that the following makes conceptual sense, although
 * it does not conform 100% with the mathematical equation:
 * - If the Power of the Original Image at any point is NULL,
 *   then G(u,v) at that point should also be NULL.  Up to this
 *   point in the code, we have used InverseFilter_Image() to determine
 *   what to do when p_orig was zero.  We will now run through the
 *   original image and set it to zero whenever the p_orig image is zero.
 */
/*
 * It is safe only to check the data type of p_orig here, since
 * we know from our cast_Image() operations that the data types
 * of all the images are identical at this point.
 */
   dtype=p_orig->image_ptr[0]->data_type;
   for(i=0; i<p_orig->bands; i++) {
      for(rmax=0; rmax<p_orig->image_ptr[i]->rows; rmax++) {
         for(cmax=0; cmax<p_orig->image_ptr[i]->cols; cmax++) {
            switch(dtype) {
               case CVIP_BYTE:
                  if(((byte *)p_orig->bandP[i]->rptr[rmax])[cmax]==0) 
                     ((byte *)retImg->bandP[i]->rptr[rmax])[cmax]=0; 
                  break;
               case CVIP_SHORT:
                  if(((short *)p_orig->bandP[i]->rptr[rmax])[cmax]==0) 
                     ((short *)retImg->bandP[i]->rptr[rmax])[cmax]=0; 
                  break;
               case CVIP_INTEGER:
                  if(((int *)p_orig->bandP[i]->rptr[rmax])[cmax]==0) 
                     ((int *)retImg->bandP[i]->rptr[rmax])[cmax]=0; 
                  break;
               case CVIP_FLOAT:
                  if(((float *)p_orig->bandP[i]->rptr[rmax])[cmax]==0) 
                     ((float *)retImg->bandP[i]->rptr[rmax])[cmax]=0; 
                  break;
               case CVIP_DOUBLE:
                  if(((double *)p_orig->bandP[i]->rptr[rmax])[cmax]==0) 
                     ((double *)retImg->bandP[i]->rptr[rmax])[cmax]=0; 
	          break;
            }
         }
      }
   }
   delete_Image(p_orig);
/*
 * Before returning, add history
 * information to the image.
 */
   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=1;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_INTEGER;
   hptr=(int *)malloc(sizeof(int));
   hptr[0]=cmax;
   stuff->packetP->dptr = (void**)malloc(sizeof(void*));
   stuff->packetP->dptr[0]=(void *)hptr;
   stuff->packetP->dsize = 1;
   stuff->next=(HISTORY)NULL;
   stuff->ftag=FFT;
   history_add(retImg, stuff);
   stuff=NULL;

   stuff=(HISTORY)malloc(sizeof(struct history));
   stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
   stuff->packetP->dsize=2;
   stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
   stuff->packetP->dtype[0]=CVIP_INTEGER;
   stuff->packetP->dtype[1]=CVIP_INTEGER;
   stuff->packetP->dsize = 2;
   stuff->packetP->dptr = (void**)malloc(sizeof(void*)*stuff->packetP->dsize);
   stuff->packetP->dptr[0]=(void *)malloc(sizeof(int));
   stuff->packetP->dptr[1]=(void *)malloc(sizeof(int));
   memcpy((void *)stuff->packetP->dptr[0], (const void *)&choice, sizeof(int));
   memcpy((void *)stuff->packetP->dptr[1], (const void *)&cutoff, sizeof(int));
   stuff->next=(HISTORY)NULL;
   stuff->ftag=WIENER_FILTER;
   history_add(retImg, stuff);
   stuff=NULL;
/*
 * Since we scaled all FFT'ed images by 1/N, well now scale back
 * the result by N.
 */
   for(i=0; i<retImg->bands; i++) {
      retImg->bandP[i]=scale_Matrix(retImg->bandP[i], rmax);
   }
   return(retImg);
}

/*
 * Routine name: parametric_wiener()
 * Description: Short and sweet setup routine which uses 
 * 		wiener() to do all the work.
 *		parametric_wiener takes the same arguments as
 *		does wiener(), and also takes a floating-
 *		point value for 'gamma'.  It then multiplies
 *		p_noise by gamma before calling wiener().
 *		The only trick here is to scale by the square root
 *		of gamma if our noise power is not in the frequency
 *		domain, since we will later square p_noise.
 *
 * Author: Arve Kjoelen, Wed Nov 20 15:36:06 CST 1996
 */
Image *parametric_wiener(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, float gamma, int choice, int cutoff) {
   int i;

   if(!history_check(FFT, p_noise))
      gamma=(float)sqrt((double)gamma);

   for(i=0; i<p_noise->bands; i++) {
      p_noise->bandP[i]=scale_Matrix((const Matrix *)p_noise->bandP[i], (double)gamma);
   }
   return(wiener(degr, degr_fn, p_noise, p_orig, choice, cutoff));
}

/*
 * Routine name: power_spect_eq()
 * Description: Setup routine which calls pse_n_gm_filter().
 * 		passes a 1 to pse_n_gm_filter, indicating that
 *		an alpha=0.5 should be assumed.
 *
 * Author: Arve Kjoelen, Wed Nov 27 10:41:00 CST 1996
 */
Image *power_spect_eq(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, int choice, int cutoff) {
   return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff, 1));
}
/*
 * Routine name: pse_n_gm_filter()
 * Description: Setup routine which uses 
 * 		wiener() to do most work.
 *		pse_n_gm_filter() takes the same arguments as
 *		does wiener().  The difference between the two
 *		is that pse_n_gm_filter uses alpha!=1, and so
 *		needs to do some extra work to compute a root
 *		of what is returned by wiener().  pse_n_gm_filter()
 *		is passed a number which indicates which root to calculate.
 *		when this number is :
 *		1 -> calculates square root, so equivalent to alpha=0.5.
 *		2 -> calculates cube root, so equivalent to alpha=1/3.
 *		3 -> equivalent to alpha=2/3.
 *		4 -> equivalent to alpha=1/9.
 *		5 -> equivalent to alpha=1/4.
 *		6 -> equivalent to alpha=1/8.
 *		7 -> equivalent to alpha=1/6.
 *		8 -> equivalent to alpha=5/6.
 *
 *	Note: pse_n_gm_filter represents Power_Spectrum_Equalization and
 *	      Geometric_Mean_Filter, since it may be used to implement both.
 *
 *
 * Author: Arve Kjoelen, Wed Nov 27 10:41:00 CST 1996
 */
Image *pse_n_gm_filter(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, int choice, int cutoff, int root) {
   int i, bsize=0;
   Matrix *tpMat;
   Image *conj_Img, *sq_Image, *retImg, *degr_save;

/*
 * First, we'll make ourselves a copy of H(u,v), since it will be
 * freed later on.  We also need to do the FFT with the correct
 * blocksize, if needed.
 */
   if(!history_check(FFT, degr_fn)) {
      bsize=MAX(degr->bandP[0]->rows, degr->bandP[0]->cols);
      bsize=MAX(bsize,degr_fn->bandP[0]->rows);
      bsize=MAX(bsize,degr_fn->bandP[0]->cols);
      bsize=MAX(bsize,p_noise->bandP[0]->rows);
      bsize=MAX(bsize,p_noise->bandP[0]->cols);
      bsize=MAX(bsize,p_orig->bandP[0]->rows);
      bsize=MAX(bsize,p_orig->bandP[0]->cols);
      /* Function to convert the block size to a power of 2 */
      bsize=block_power_2(bsize);
      sq_Image=duplicate_Image(degr_fn);
      sq_Image=fft_transform(sq_Image, bsize);
   }
/*
 * Call the wiener, after first making a copy of degr.
 */
   degr_save=duplicate_Image(degr);
   if(!history_check(FFT, degr)) {
      degr_save=fft_transform(degr_save, bsize);
   }
   retImg=wiener(degr, degr_fn, p_noise, p_orig, choice, cutoff);

/*
 * calculate H*(u,v) and [H(u,v) squared].
 * In addition, scale our FFT'ed image by 1/N, since the
 * definition uses 1/(N*N) in the forward, while our code uses 1/N.
 */
   conj_Img=duplicate_Image(sq_Image);
   for(i=0; i<sq_Image->bands; i++) {
      sq_Image->bandP[i]=scale_Matrix(sq_Image->bandP[i], (double)(1.0/bsize));
      conj_Img->bandP[i]=scale_Matrix(conj_Img->bandP[i], (double)(1.0/bsize));
      (void)conj_Matrix(conj_Img->bandP[i]);
      tpMat=sq_Image->bandP[i];
      sq_Image->bandP[i]=square_mag_Matrix(sq_Image->bandP[i]);
      delete_Matrix(tpMat);
   }
/*
 * reuse some variables, since images have all been freed by wiener()
 */
   (void)inverseFilter_Image(&retImg, conj_Img, sq_Image, (float)cutoff, choice);
   delete_Image(conj_Img);
   delete_Image(sq_Image);

/*
 * If we compare our current status with our equation, the two
 * main components of the equation are contained in degr and retImg.
 * We need to multiply these, before computing the root of the
 * result.
 */
   for(i=0; i<degr_save->bands; i++) {
      tpMat=retImg->bandP[i];
      retImg->bandP[i]=multPWise_Matrix(retImg->bandP[i], degr_save->bandP[i]);
      delete_Matrix(tpMat);
   }
   delete_Image(degr_save);
/*
 * Compute the root and scale back by N.
 */
   fprintf(stderr, "Computing result to the alpha power\n");
   if(root>7) {
      degr=duplicate_Image(retImg);
   }
   for(i=0; i<retImg->bands; i++) {
      if(root==1) /* alpha=1/2 */
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
      else if(root==2) /* alpha=1/3 */
         retImg->bandP[i]=cbrt_Matrix(retImg->bandP[i]);
      else if(root==3) { /* alpha=2/3 */
         retImg->bandP[i]=cbrt_Matrix(retImg->bandP[i]);
	 tpMat=retImg->bandP[i];
         retImg->bandP[i]=multPWise_Matrix(retImg->bandP[i], retImg->bandP[i]);
         delete_Matrix(tpMat);
      }
      else if(root==4) { /* alpha=1/9 */
         retImg->bandP[i]=cbrt_Matrix(retImg->bandP[i]);
         retImg->bandP[i]=cbrt_Matrix(retImg->bandP[i]);
      }
      else if(root==5) { /* alpha=1/4 */
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
      }
      else if(root==6) { /* alpha=1/8 */
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
      }
      else if(root==7) { /* alpha=1/6 */
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
         retImg->bandP[i]=cbrt_Matrix(retImg->bandP[i]);
      }
      else if(root==8) { /* alpha=5/6 */
         degr->bandP[i]=cbrt_Matrix(degr->bandP[i]);
         retImg->bandP[i]=sqrt_Matrix(retImg->bandP[i]);
	 tpMat=retImg->bandP[i];
         retImg->bandP[i]=multPWise_Matrix(retImg->bandP[i], degr->bandP[i]);
         delete_Matrix(tpMat);
      }
      retImg->bandP[i]=scale_Matrix(retImg->bandP[i], bsize);
   }
   if(root>7) {
      delete_Image(degr);
   }
   write_Image(retImg, "/tmp/img1", CVIP_YES, CVIP_NO, VIP, CVIP_YES);
   return(retImg);
}


/*
 * Routine name: geometric_mean()
 * Description: Setup routine which uses 
 * 		wiener() to do most work.
 *		geometric_mean takes the same arguments as
 *		does wiener(), and also takes a floating-
 *		point value for 'gamma'.  It then multiplies
 *		p_noise by gamma before calling wiener().
 *		The only trick here is to scale by the square root
 *		of gamma if our noise power is not in the frequency
 *		domain, since we will later square p_noise.
 *		Also note that power_spect_eq() assumes alpha=0.5,
 *		and needs to do some extra work to compute the square root
 *		of what is returned by wiener().  For now, the
 *		value of alpha is fixed at 0.5, since it will take some time
 *		to code a routine that computes an arbitrary root of
 *		a complex number.
 *
 * Author: Arve Kjoelen, Wed Nov 27 10:41:00 CST 1996
 */
Image *geometric_mean(Image *degr, Image *degr_fn, Image *p_noise, Image *p_orig, float gamma, float alpha, int choice, int cutoff) {
   int i;

/*
 * Not all alpha are supported, so here are our special cases:
 * - for alpha=0, the equation is reduced to a parametric wiener.
 * - for alpha=1/3, we can use the power spectrum equalization filter,
 *   although we must make sure to multiply p_noise by gamma before
 *   calling pse_n_gm_filter.  We must also pass the value 2 as our
 *   last argument to that function in order to have cuber roots calculated.
 * - for alpha=0.5, we can use the power spectrum equalization filter,
 *   although we must make sure to multiply p_noise by gamma before
 *   calling pse_n_gm_filter.  We pass the value 1 as our final
 *   argument to that function to indicate that we want the square root
 *   computed.
 * - for alpha=1, the equation reduces to an inverse filter.
 */
   if(CVIP_NEAR(alpha, 0.0)) {
      return(parametric_wiener(degr, degr_fn, p_noise, p_orig, gamma, choice, cutoff));
   }
   else if(CVIP_NEAR(alpha, 1.0)) {
      delete_Image(p_noise);
      delete_Image(p_orig);
      return(inverse_xformfilter(degr, degr_fn,choice,(float)cutoff));
   }
/*
 * alpha is now some funky number.
 */
   if(!history_check(FFT, p_noise))
      gamma=(float)sqrt((double)gamma);

   for(i=0; i<p_noise->bands; i++) 
      p_noise->bandP[i]=scale_Matrix((const Matrix *)p_noise->bandP[i], (double)gamma);

   if(CVIP_NEAR(alpha, 0.111111))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,4));
   else if(CVIP_NEAR(alpha, 0.1250))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,6));
   else if(CVIP_NEAR(alpha, 0.166667))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,7));
   else if(CVIP_NEAR(alpha, 0.2500))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,5));
   else if(CVIP_NEAR(alpha, 0.333333))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,2));
   else if(CVIP_NEAR(alpha, 0.5))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,1));
   else if(CVIP_NEAR(alpha, 0.666667))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,3));
   else if(CVIP_NEAR(alpha, 0.833333))
      return(pse_n_gm_filter(degr, degr_fn, p_noise, p_orig, choice, cutoff,8));
   else {
      fprintf(stderr, "Unsupported value for alpha: %f\n", alpha);
      return(degr);
   }
}


/*
 * Routine name: idealFilter_Matrix.
 *
 * Description: Creates a "filter matrix" - a matrix of selectable
 * size consisting of 1s and 0s - the zeros are all located outside
 * a user-selectable radius.
 *
 * Credits: Greg Hance, Arve Kjoelen
 */
static int idealFilter_Matrix(
	Matrix 	**idealP, 
	float 	fc, 
	int 	nr,
	int 	nc
)
{
	byte	**dataP;
	int 	u, un, v, vn, n2;
	float	D;
	
	*idealP = new_Matrix(nr, nc, CVIP_BYTE, REAL);
	dataP = getData_Matrix(*idealP);

	n2 = nr/2;
	for(u = 0 ; u < nr ; u++) {
		for(v = 0 ; v < nc ; v++) {
			un = u - n2;
			vn = v - n2;
			D = sqrt(un*un + vn*vn);
			if(D <= fc)
				dataP[u][v] = 1;
		}
	}
	
	return 1;
}
/*
 * Routine Name: inverse_xformfilter
 * 
 * Description: Setup function for Inverse Filter.  Checks
 * input images and performs FFT on them if necessary, automatically
 * extending their sizes (zero-padding). Asks the user how to treat
 * occurences of 0 in the denominator.  Also asks whether optional
 * 'filtering' should be done, wherein F(u,v) is set to 0 when
 * sqrt(u*u + v*v) is outside a certain radius. 
 *
 * The function is also a setup function for a simplified Wiener
 * filter, in which the ratio of the noise power to the original
 * image power is approximated by a constant K.  If the Wiener filter
 * is desired, the function is passed a 1 for the parameter xform.
 *
 * Parameters passed: 	imgP, which will hold the result.
 *			numP, the numerator.
 *			denP, the denominator.
 *
 * History information for inverse filter:
 * parameter #1 = filter cutoff.
 * parameter #2 = method (1: F(u,v)=G(u,v) when H(u,v)=0
 *			  2: F(u,v)=0      when H(u,v)=0)
 */

Image *inverse_xformfilter(
	Image	*numP,
	Image	*denP,
	int choice,
	float cutoff
)
{
	int		block_size=0, *hptr, band;
	Image 		*imgP;
	HISTORY		stuff;

        block_size=MAX(numP->bandP[0]->rows,numP->bandP[0]->cols);
	block_size=MAX(block_size,MAX(denP->bandP[0]->rows,denP->bandP[0]->cols));
	/* Function to convert the block size to a power of 2 */
        block_size=block_power_2(block_size);
   if(!history_check(FFT, numP)) {
       fprintf(stdout,  "The degraded image appears to be in the spatial domain\n");
       fprintf(stdout,  "Performing FFT and any necessary zero-padding...\n");
       numP=fft_transform(numP, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(numP, stuff);
       stuff=NULL;
   }
   if(!history_check(FFT, denP)) {
       fprintf(stdout, "The degradation function appears to be in the spatial domain\n");
       fprintf(stdout, "Performing FFT and any necessary zero-padding...\n");
       denP=fft_transform(denP, block_size);
       stuff=(HISTORY)malloc(sizeof(struct history));
       stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
       stuff->packetP->dsize=1;
       stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE));
       stuff->packetP->dtype[0]=CVIP_INTEGER;
       hptr=(int *)malloc(sizeof(int));
       hptr[0]=block_size;
       stuff->packetP->dptr = (void**)malloc(sizeof(void*));
       stuff->packetP->dptr[0]=(void *)hptr;
       stuff->packetP->dsize = 1;
       stuff->next=(HISTORY)NULL;
       stuff->ftag=FFT;
       history_add(denP, stuff);
       stuff=NULL;
   }


/*
 * Divide by N, since our definition of the FFT
 * uses 1/N instead of 1/(N*N) as the forward multiplier.
 * this may seem unnecessary, but must be done whenever we
 * choose: '(1) Let F(u,v) = G(u,v)' from above.
 */
    for(band=0; band<numP->bands; band++) {
       numP->image_ptr[band]=scale_Matrix(numP->image_ptr[band], 1.0/(double)numP->image_ptr[band]->cols);
       denP->image_ptr[band]=scale_Matrix(denP->image_ptr[band], 1.0/(double)numP->image_ptr[band]->cols);
     }
/*
 * Perform the Inverse Filter.
 */
    (void)inverseFilter_Image(&imgP, numP, denP, (float)cutoff, choice);
/*
 * scale back to avoid spatial domain underflow (although unlikely).
 */
    for(band=0; band<numP->bands; band++) {
       imgP->image_ptr[band]=scale_Matrix((const Matrix *)imgP->image_ptr[band], (double)imgP->image_ptr[band]->cols);
    }

	stuff=(HISTORY)malloc(sizeof(struct history));
	stuff->packetP = (PACKET *)malloc(sizeof(PACKET));
	stuff->ftag = INVERSEFILTER;
	stuff->packetP->dsize=2;
	stuff->packetP->dtype=(CVIP_TYPE *)malloc(sizeof(CVIP_TYPE)*stuff->packetP->dsize);
	stuff->packetP->dtype[0]=CVIP_INTEGER;
	stuff->packetP->dtype[1]=CVIP_INTEGER;
	stuff->packetP->dptr = (void**)malloc(sizeof(void *)*stuff->packetP->dsize);
	stuff->packetP->dptr[0] = (void *)malloc(sizeof(int));
	stuff->packetP->dptr[1] = (void *)malloc(sizeof(int));
	memcpy((void *)stuff->packetP->dptr[0], (const void *)&choice, sizeof(int));
	memcpy((void *)stuff->packetP->dptr[1], (const void *)&cutoff, sizeof(int));
	stuff->next=(HISTORY)NULL;
	history_add(imgP, stuff);
/*
 * If block_size is not 0, we did the FFT in
 * this module, and need to free numP, since otherwise, it will
 * cause a memory leak.
 */
	delete_Image(numP);
	delete_Image(denP);
	return(imgP);
}






/*
 * Routine Name: inverseFilter_Image
 *
 * Description: Performs the inverse filter.  Does not explicitly
 * return the result, but places it inside the image passed.
 * 
 */ 
int
inverseFilter_Image(
	Image		**imgP,
	Image	*numP,
	Image	*denP,
	float		cutoff,
	int		choice
)
{
	
	Matrix		*divP, *ideP, *filP, *masP;
	unsigned	bands, nr, nc;
	register int	i, j;
	void		*dataP;
	double		val;
	

	bands 	= getNoOfBands_Image(numP);
	nr	= getNoOfRows_Image(numP);
	nc	= getNoOfCols_Image(numP);

	*imgP = duplicate_Image(numP);

	if(cutoff)
		(void) idealFilter_Matrix(&ideP, cutoff, nr, nc);

	masP = 	new_Matrix(nr, nc, getDataType_Image(numP),
		getDataFormat_Image(numP));

	dataP = getRealData_Matrix(masP);

	if(choice == 1)
		val = 1.0;
	else
		val = 0.0;

	for(i=0; i < nr; i++)
		for(j=0; j < nc; j++)
			switch(getDataType_Matrix(masP)) {
				case CVIP_BYTE:
					((byte **) dataP)[i][j] = (byte) val;
					break;
				case CVIP_FLOAT:
					((float **) dataP)[i][j] = (float) val;
					break;
				case CVIP_DOUBLE:
					((double **) dataP)[i][j] = val;
					break;
			}
	

	for(i=0; i < bands; i++) {
		delete_Matrix(getBand_Image(*imgP,i));
		if(cutoff)
			filP = multPWise_Matrix(getBand_Image(denP,i), ideP);
		else
			filP = getBand_Image(denP,i);
		(void) 	divide_Matrix(&divP, 
			getBand_Image(numP,i),
			filP,
			masP );
		if(cutoff)
			delete_Matrix(filP);
		else
			filP = (NULL);

		setBand_Image(*imgP,divP,i);
		/*delete_Matrix(divP);*/
	}

	delete_Matrix(masP);
	if(cutoff)
		delete_Matrix(ideP);

	return 1;
}

int block_power_2(block_size)
{
	/* temporary variables for converting the block size to a power of 2 */
  int temp=1,temp1;

/* code for converting the blocksize to a power of 2 */
  for(temp1=1;temp<block_size;temp1++)
  {
        temp=(int)(pow(2.0,(double)temp1));
        if(block_size==temp)
        {
                temp1++;
                break;
        }
  }
  temp1--;
  block_size=(int)(pow(2.0,(double)temp1));
  return(block_size);
}
